"""
Several implementation details are derived from the scipy csd and welch
implementations. For more info, see

https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.welch.html

and

https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.csd.html
"""

from typing import Optional, Union

import torch
from jaxtyping import Float
from torch import Tensor

from .types import (
    FrequencySeries1to3d,
    PSDTensor,
    TimeSeries1to3d,
    WaveformTensor,
)


def median(x: Float[Tensor, "... size"], axis: int) -> Float[Tensor, "..."]:
    """
    Implements a median calculation that matches numpy's
    behavior for an even number of elements and includes
    the same bias correction used by scipy's implementation.
    see https://github.com/scipy/scipy/blob/main/scipy/signal/_spectral_py.py#L2066 # noqa
    """
    n = x.shape[axis]
    ii_2 = 2 * torch.arange(1.0, (n - 1) // 2 + 1)
    bias = 1 + torch.sum(1.0 / (ii_2 + 1) - 1.0 / ii_2)
    return torch.quantile(x, q=0.5, axis=axis) / bias


def _validate_shapes(
    x: Tensor, nperseg: int, y: Optional[Tensor] = None
) -> None:
    if x.shape[-1] < nperseg:
        raise ValueError(
            "Number of samples {} in input x is insufficient "
            "for number of fft samples {}".format(x.shape[-1], nperseg)
        )
    elif x.ndim > 3:
        raise ValueError(
            f"Can't compute spectral density on tensor with shape {x.shape}"
        )

    if y is None:
        return

    # acceptable combinations of shapes:
    # x: time, y: time
    # x: channel x time, y: time OR channel x time
    # x: batch x channel x time, y: batch x channel x time OR batch x time
    if x.shape[-1] != y.shape[-1]:
        raise ValueError(
            "Time dimensions of x and y tensors must "
            "be the same, found {} and {}".format(x.shape[-1], y.shape[-1])
        )
    elif x.ndim == 1 and not y.ndim == 1:
        raise ValueError(
            "Can't compute cross spectral density of "
            "1D tensor x with {}D tensor y".format(y.ndim)
        )
    elif x.ndim > 1 and y.ndim == x.ndim:
        if not y.shape == x.shape:
            raise ValueError(
                "If x and y tensors have the same number "
                "of dimensions, shapes must fully match. "
                "Found shapes {} and {}".format(x.shape, y.shape)
            )
    elif x.ndim > 1 and y.ndim != (x.ndim - 1):
        raise ValueError(
            "Can't compute cross spectral density of "
            "tensors with shapes {} and {}".format(x.shape, y.shape)
        )
    elif x.ndim > 2 and y.shape[0] != x.shape[0]:
        raise ValueError(
            "If x is a 3D tensor and y is a 2D tensor, "
            "0th batch dimensions must match, but found "
            "values {} and {}".format(x.shape[0], y.shape[0])
        )


def fast_spectral_density(
    x: TimeSeries1to3d,
    nperseg: int,
    nstride: int,
    window: Float[Tensor, " {nperseg//2+1}"],
    scale: float,
    average: str = "median",
    y: Optional[TimeSeries1to3d] = None,
) -> FrequencySeries1to3d:
    """
    Compute the power spectral density of a multichannel
    timeseries or a batch of multichannel timeseries, or
    the cross power spectral density of two such timeseries.
    This implementation is non-exact for the two lowest
    frequency bins, since it implements centering using the
    mean for the entire timeseries rather than on a per-
    window basis. The benefit of this is a faster implementation,
    which might be beneficial for cases where the lowest
    frequency components are going to be discarded anyway.

    Args:
        x:
            The timeseries tensor whose power spectral density
            to compute, or for cross spectral density the
            timeseries whose fft will be conjugated. Can have
            shape `(batch_size, num_channels, length * sample_rate)`,
            `(num_channels, length * sample_rate)`, or
            `(length * sample_rate)`.
        nperseg:
            Number of samples included in each FFT window
        nstride:
            Stride between FFT windows
        window:
            Window array to multiply by each FFT window before
            FFT computation. Should have length `nperseg // 2 + 1`.
        scale:
            Scale factor to multiply the FFT'd data by, related to
            desired units for output tensor (e.g. letting this equal
            `1 / (sample_rate * (window**2).sum())` will give output
            units of density, $\\text{Hz}^-1$$.
        average:
            How to aggregate the contributions of each FFT window to
            the spectral density. Allowed options are `'mean'` and
            `'median'`.
        y:
            Timeseries tensor to compute cross spectral density
            with `x`. If left as `None`, `x`'s power spectral
            density will be returned. Otherwise, if `x` is 1D,
            `y` must also be 1D. If `x` is 2D, the assumption
            is that this represents a single multi-channel timeseries,
            and `y` must be either 2D or 1D. In the former case,
            the cross-spectral densities of each channel will be
            computed individually, so `y` must have the same shape as `x`.
            Otherwise, this will compute the CSD of each of `x`'s channels
            with `y`. If `x` is 3D, this will be assumed to be a batch
            of multi-channel timeseries. In this case, `y` can either
            be 3D, in which case each channel of each batch element will
            have its CSD calculated or 2D, which has two different options.
            If `y`'s 0th dimension matches `x`'s 0th dimension, it will
            be assumed that `y` represents a batch of 1D timeseries, and
            for each batch element this timeseries will have its CSD with
            each channel of the corresponding batch element of `x`
            calculated. Otherwise, it sill be assumed that `y` represents
            a single multi-channel timeseries, in which case each channel
            of `y` will have its CSD calculated with the corresponding
            channel in `x` across _all_ of `x`'s batch elements.
    Returns:
        Tensor of power spectral densities of `x` or its cross spectral
        density with the timeseries in `y`.
    """

    _validate_shapes(x, nperseg, y)

    if x.ndim > 2:
        # stft only works on 2D input, so roll the
        # channel dimension out along the batch
        batch_size = x.shape[0]
        num_channels = x.shape[1]
        x = x.reshape(-1, x.shape[-1])

        if y is not None and y.ndim == 3:
            # do the same thing for y if necessary
            y = y.reshape(-1, x.shape[-1])
    else:
        # this is just a multichannel timeseries, so
        # we can ignore batch reshapes later
        batch_size = None

    x = x - x.mean(axis=-1, keepdims=True)
    fft = torch.stft(
        x,
        n_fft=nperseg,
        hop_length=nstride,
        window=window,
        normalized=False,
        center=False,
        return_complex=True,
    )

    other = fft
    if y is not None:
        y = y - y.mean(axis=-1, keepdims=True)
        y_fft = torch.stft(
            y,
            n_fft=nperseg,
            hop_length=nstride,
            window=window,
            normalized=False,
            center=False,
            return_complex=True,
        )
        if batch_size is not None and fft.shape[0] != y_fft.shape[0]:
            # x is batched but y is not batched, so expand x's 0th
            # dimension back out so that we can multiply these arrays
            nfreq = nperseg // 2 + 1
            fft = fft.reshape(batch_size, num_channels, nfreq, -1)
            y_fft = y_fft[:, None]
        other = y_fft

    fft = torch.conj(fft) * other
    if y is None:
        fft = fft.real

    # some overly complex logic to find the frequency dimension
    # to apply this inscrutable signal processing step of
    # multiplying the non-edge frequency bins by 2
    # TODO: move this after aggregation since it won't affect
    # output value, and we have one less axis to do it over?
    stop = None if nperseg % 2 else -1
    if x.ndim == 1:
        fft[1:stop] *= 2
    elif fft.ndim < 4:
        fft[:, 1:stop] *= 2
    else:
        fft[:, :, 1:stop] *= 2
    fft *= scale

    if average == "mean":
        fft = fft.mean(axis=-1)
    else:
        # if this is a cross spectral density, fft will
        # be complex and so we'll compute the median along
        # both axes
        if y is not None:
            real_median = median(fft.real, -1)
            imag_median = 1j * median(fft.imag, -1)
            fft = real_median + imag_median
        else:
            fft = median(fft, -1)

    if fft.ndim == 2 and batch_size is not None:
        # if we still haven't expanded the batch dimension
        # back out, do so now
        fft = fft.reshape(batch_size, num_channels, -1)
    return fft


def spectral_density(
    x: TimeSeries1to3d,
    nperseg: int,
    nstride: int,
    window: Float[Tensor, " {nperseg//2+1}"],
    scale: float,
    average: str = "median",
) -> FrequencySeries1to3d:
    """
    Compute the power spectral density of a multichannel
    timeseries or a batch of multichannel timeseries.
    This implementation is exact for all frequency bins, but
    slower than the fast implementation.

    Args:
        x:
            The timeseries tensor whose power spectral density
            to compute, or for cross spectral density the
            timeseries whose fft will be conjugated. Can have
            shape `(batch_size, num_channels, length * sample_rate)`,
            `(num_channels, length * sample_rate)`, or
            `(length * sample_rate)`.
        nperseg:
            Number of samples included in each FFT window
        nstride:
            Stride between FFT windows
        window:
            Window array to multiply by each FFT window before
            FFT computation. Should have length `nperseg // 2 + 1`.
        scale:
            Scale factor to multiply the FFT'd data by, related to
            desired units for output tensor (e.g. letting this equal
            `1 / (sample_rate * (window**2).sum())` will give output
            units of density, $\\text{Hz}^-1$$.
        average:
            How to aggregate the contributions of each FFT window to
            the spectral density. Allowed options are `'mean'` and
            `'median'`.
    """

    _validate_shapes(x, nperseg)

    # for non-fast implementation, we need to unfold
    # the tensor along the time dimension ourselves
    # to detrend each segment individually, so start
    # by converting x to a 4D tensor so we can use
    # torch's Unfold op
    if x.ndim == 1:
        reshape = []
        x = x[None, None, None, :]
    elif x.ndim == 2:
        reshape = [len(x)]
        x = x[None, :, None, :]
    elif x.ndim == 3:
        reshape = list(x.shape[:-1])
        x = x[:, :, None, :]

    # calculate the number of segments and trim x along
    # the time dimensions so that we can unfold it exactly
    num_segments = (x.shape[-1] - nperseg) // nstride + 1
    stop = (num_segments - 1) * nstride + nperseg
    x = x[..., :stop]

    # unfold x into overlapping segments and detrend and window
    # each one individually before computing the rfft. Unfold
    # will produce a batch x (num_channels * num_segments) x nperseg
    # shaped tensor
    unfold_op = torch.nn.Unfold((1, num_segments), dilation=(1, nstride))
    x = unfold_op(x)
    x = x - x.mean(axis=-1, keepdims=True)
    x *= window

    # after the fft, we'll have a
    # batch x (num_channels * num_segments) x nfreq
    # sized tensor
    fft = torch.fft.rfft(x, axis=-1).abs() ** 2

    if nperseg % 2:
        fft[..., 1:] *= 2
    else:
        fft[..., 1:-1] *= 2
    fft *= scale

    # unfold the batch and channel dimensions back
    # out if there were any to begin with, putting
    # the segment dimension as the second to last
    reshape += [num_segments, -1]
    fft = fft.reshape(*reshape)

    if average == "mean":
        return fft.mean(axis=-2)
    else:
        return median(fft, -2)


def truncate_inverse_power_spectrum(
    psd: PSDTensor,
    fduration: Union[Float[Tensor, " time"], float],
    sample_rate: float,
    highpass: Optional[float] = None,
    lowpass: Optional[float] = None,
) -> PSDTensor:
    """
    Truncate the length of the time domain response
    of a whitening filter built using the specified
    `psd` so that it has maximum length `fduration`
    seconds. This is meant to mitigate the impact
    of sharp features in the background PSD causing
    time domain responses longer than the segments
    to which the whitening filter will be applied.

    Implementation details adapted from
    https://github.com/vivinousi/gw-detection-deep-learning/blob/203966cc2ee47c32c292be000fb009a16824b7d9/modules/whiten.py#L8  # noqa

    Args:
        psd:
            The one-sided power spectraul density used
            to construct a whitening filter.
        fduration:
            Desired length in seconds of the time domain
            response of a whitening filter built using
            this PSD, or a window of this length to taper
            the edges of the time domain response of the
            filter. If passed as a float, a Hann window
            of this length will be used.
        sample_rate:
            Rate at which the time domain data to which
            the whitening filter will be applied has been
            sampled.
        highpass:
            If specified, will zero out the frequency response
            of all frequencies below this value in Hz. If left
            as `None`, no highpass filtering will be applied.
        lowpass:
            If specified, will zero out the frequency response
            of all frequencies above this value in Hz. If left
            as `None`, no lowpass filtering will be applied.
    Returns:
        The PSD with its time domain response truncated
            to `fduration` and any filtered frequencies
            tapered.
    """

    num_freqs = psd.size(-1)
    N = (num_freqs - 1) * 2

    # use the inverse of the ASD as the
    # impulse response function
    inv_asd = 1 / psd**0.5

    # zero out frequencies if we want the filter
    # to perform highpass/lowpass filtering
    df = sample_rate / N
    if highpass is not None:
        idx = int(highpass / df)
        inv_asd[:, :, :idx] = 0
    if lowpass is not None:
        idx = int(lowpass / df)
        inv_asd[:, :, idx:] = 0

    if inv_asd.size(-1) % 2:
        inv_asd[:, :, -1] = 0

    # now convert to time domain representation
    q = torch.fft.irfft(inv_asd, n=N, norm="forward", dim=-1)

    # taper the edges of the TD filter
    if isinstance(fduration, Tensor):
        pad = fduration.size(-1) // 2
        window = fduration
    else:
        pad = int(fduration * sample_rate / 2)
        window = torch.hann_window(2 * pad, dtype=torch.float64)
        window = window.to(q.device)

    # 0 out anything else between the tapering regions
    q[:, :, :pad] *= window[-pad:]
    q[:, :, -pad:] *= window[:pad]
    if 2 * pad < q.size(-1):
        q[:, :, pad : q.size(-1) - pad] = 0

    # convert back to the frequency domain
    # to build the desired PSD
    inv_asd = torch.fft.rfft(q, n=N, norm="forward", dim=-1)
    inv_psd = inv_asd * inv_asd.conj()
    psd = 1 / inv_psd.abs()
    return psd / 2


def normalize_by_psd(
    X: WaveformTensor,
    psd: PSDTensor,
    sample_rate: float,
    pad: int,
):
    # compute the FFT of the section we want to whiten
    # and divide it by the ASD of the background section.
    # If the ASD of any background bin hit inf, set the
    # corresponding bin to 0
    X = X - X.mean(-1, keepdims=True)
    X_tilde = torch.fft.rfft(X.double(), norm="forward", dim=-1)
    X_tilde = X_tilde / psd**0.5
    X_tilde[torch.isnan(X_tilde)] = 0

    # convert back to the time domain and normalize
    # TODO: what's this normalization factor?
    X = torch.fft.irfft(X_tilde, n=X.shape[-1], norm="forward", dim=-1)
    X = X.float() / sample_rate**0.5

    # slice off corrupted data at edges of kernel
    X = X[:, :, pad:-pad]
    return X


def whiten(
    X: WaveformTensor,
    psd: PSDTensor,
    fduration: Union[Float[Tensor, " time"], float],
    sample_rate: float,
    highpass: Optional[float] = None,
    lowpass: Optional[float] = None,
) -> WaveformTensor:
    """
    Whiten a batch of timeseries using the specified
    background one-sided power spectral densities (PSDs),
    modified to have the desired time domain response length
    `fduration` and possibly to highpass/lowpass filter.

    Args:
        X:
            batch of multichannel timeseries to whiten
        psd:
            PSDs use to whiten the data. The frequency
            response of the whitening filter will be roughly
            the inverse of the square root of this PSD, ensuring
            that data from the same distribution will have
            approximately uniform power after whitening.
            If 2D, each batch element in `X` will be whitened
            using the same PSDs. If 3D, each batch element will
            be whitened by the PSDs contained along the 0th
            dimenion of `psd`, and so the first two dimensions
            of `X` and `psd` should match.
        fduration:
            Desired length in seconds of the time domain
            response of a whitening filter built using
            this PSD, or a window of this length to taper
            the edges of the time domain response of the
            filter. If passed as a float, a Hann window
            of this length will be used. Moreover, half of
            this length will be removed from each edge of
            the whitened timeseries to account for filter
            settle-in time.
        sample_rate:
            Rate at which the data in `X` has been sampled
        highpass:
            The frequency in Hz at which to highpass filter
            the data, setting the frequency response in the
            whitening filter to 0. If left as `None`, no
            highpass filtering will be applied.
        lowpass:
            The frequency in Hz at which to lowpass filter
            the data, setting the frequency response in the
            whitening filter to 0. If left as `None`, no
            lowpass filtering will be applied.
    Returns:
        Batch of whitened multichannel timeseries with
            `fduration / 2` seconds trimmed from each side.
    """

    # figure out how much data we'll need to slice
    # off after whitening
    if isinstance(fduration, Tensor):
        pad = fduration.size(-1) // 2
    else:
        pad = int(fduration * sample_rate / 2)

    N = X.size(-1)
    if N <= (2 * pad):
        raise ValueError(
            "Not enough timeseries samples {} for "
            "number of padded samples {}".format(N, 2 * pad)
        )

    # normalize the number of expected dimensions in the PSD
    while psd.ndim < 3:
        psd = psd[None]

    # possibly interpolate our PSD to match the number
    # of frequency bins we expect to get from X
    num_freqs = N // 2 + 1
    if psd.size(-1) != num_freqs:
        # TODO: does there need to be any rescaling to
        # keep the integral of the PSD constant?
        psd = torch.nn.functional.interpolate(
            psd, size=(num_freqs,), mode="linear"
        )

    # truncate it to have the desired
    # time domain response length
    psd = truncate_inverse_power_spectrum(
        psd,
        fduration,
        sample_rate,
        highpass,
        lowpass,
    )

    return normalize_by_psd(X, psd, sample_rate, pad)



import torch
from torch import Tensor

from .utils import match_size


class SkipConnection(torch.nn.Module):
    def forward(self, X: Tensor, state: Tensor) -> Tensor:
        return match_size(X, state.size(-1))

    def get_out_channels(self, in_channels: int) -> int:
        return in_channels


class AddSkipConnect(SkipConnection):
    def forward(self, X: Tensor, state: Tensor) -> Tensor:
        X = super().forward(X, state)
        return X + state


class ConcatSkipConnect(SkipConnection):
    def __init__(self, groups: int = 1) -> None:
        super().__init__()
        self.groups = groups

    def get_out_channels(self, in_channels: int) -> int:
        return 2 * in_channels

    def forward(self, X: Tensor, state: Tensor) -> Tensor:
        X = super().forward(X, state)
        if self.groups == 1:
            return torch.cat([X, state], dim=1)

        num_channels = X.size(1)
        rem = num_channels % self.groups
        if rem:
            raise ValueError(
                "Number of channels in input tensor {} cannot "
                "be divided evenly into {} groups".format(
                    num_channels, self.groups
                )
            )

        X = torch.split(X, self.groups, dim=1)
        state = torch.split(state, self.groups, dim=1)
        frags = [i for j in zip(X, state) for i in j]
        return torch.cat(frags, dim=1)



from typing import List, Optional

import torch
from jaxtyping import Float
from torch import Tensor

from .. import gw
from ..types import BatchTensor


# TODO: should these live in ml4gw.waveforms submodule?
# TODO: what in here should be stored as buffers?
class WaveformSampler(torch.nn.Module):
    def __init__(
        self,
        parameters: Optional[Float[Tensor, "batch num_params"]] = None,
        **polarizations: Float[Tensor, "batch time"],
    ):
        super().__init__()
        # make sure we have the same number of waveforms
        # for all the different polarizations
        num_waveforms = None
        self.polarizations = {}
        for polarization, tensor in polarizations.items():
            if num_waveforms is not None and len(tensor) != num_waveforms:
                raise ValueError(
                    "Polarization {} has {} waveforms "
                    "associated with it, expected {}".format(
                        polarization, len(tensor), num_waveforms
                    )
                )
            elif num_waveforms is None:
                num_waveforms = tensor.shape[0]

            self.polarizations[polarization] = Tensor(tensor)

        if parameters is not None and len(parameters) != num_waveforms:
            raise ValueError(
                "Waveform parameters has {} waveforms "
                "associated with it, expected {}".format(
                    len(parameters), num_waveforms
                )
            )
        self.num_waveforms = num_waveforms
        self.parameters = parameters

    def forward(self, N: int):
        # TODO: should we allow sampling with replacement?
        if N > self.num_waveforms:
            raise ValueError(
                "Requested {} waveforms, but only {} are available".format(
                    N, self.num_waveforms
                )
            )
        # TODO: do we still really want this behavior here when a
        # user can do this without instantiating a WaveformSampler?
        elif N == -1:
            idx = torch.arange(self.num_waveforms)
            N = self.num_waveforms
        else:
            idx = torch.randint(self.num_waveforms, size=(N,))

        waveforms = {k: v[idx] for k, v in self.polarizations.items()}
        if self.parameters is not None:
            return waveforms, self.parameters[idx]
        return waveforms


class WaveformProjector(torch.nn.Module):
    def __init__(self, ifos: List[str], sample_rate: float):
        super().__init__()
        tensors, vertices = gw.get_ifo_geometry(*ifos)
        self.sample_rate = sample_rate
        self.register_buffer("tensors", tensors)
        self.register_buffer("vertices", vertices)

    def forward(
        self,
        dec: BatchTensor,
        psi: BatchTensor,
        phi: BatchTensor,
        **polarizations: Float[Tensor, "batch time"],
    ):
        ifo_responses = gw.compute_observed_strain(
            dec,
            psi,
            phi,
            detector_tensors=self.tensors,
            detector_vertices=self.vertices,
            sample_rate=self.sample_rate,
            **polarizations,
        )
        return ifo_responses



from .iirfilter import IIRFilter
from .pearson import ShiftedPearsonCorrelation
from .qtransform import QScan, SingleQTransform
from .scaler import ChannelWiseScaler
from .snr_rescaler import SnrRescaler
from .spectral import SpectralDensity
from .spectrogram import MultiResolutionSpectrogram
from .spline_interpolation import SplineInterpolate
from .waveforms import WaveformProjector, WaveformSampler
from .whitening import FixedWhiten, Whiten



import torch

from ..constants import MTSUN_SI, PI
from ..types import BatchTensor


def rotate_z(angle: BatchTensor, x, y, z):
    x_tmp = x * torch.cos(angle) - y * torch.sin(angle)
    y_tmp = x * torch.sin(angle) + y * torch.cos(angle)
    return x_tmp, y_tmp, z


def rotate_y(angle, x, y, z):
    x_tmp = x * torch.cos(angle) + z * torch.sin(angle)
    z_tmp = -x * torch.sin(angle) + z * torch.cos(angle)
    return x_tmp, y, z_tmp


def XLALSimInspiralLN(
    total_mass: BatchTensor, eta: BatchTensor, v: BatchTensor
):
    """
    See https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L2173 # noqa
    """
    return total_mass**2 * eta / v


def XLALSimInspiralL_2PN(eta: BatchTensor):
    """
    See https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L2181 # noqa
    """
    return 1.5 + eta / 6.0


def chirp_mass_and_mass_ratio_to_components(
    chirp_mass: BatchTensor, mass_ratio: BatchTensor
):
    """
    Compute component masses from chirp mass and mass ratio.
    Args:
        chirp_mass: Tensor of chirp mass values
        mass_ratio:
            Tensor of mass ratio values, `m2 / m1`,
            where m1 >= m2, so that mass_ratio <= 1
    """
    total_mass = chirp_mass * (1 + mass_ratio) ** 1.2 / mass_ratio**0.6
    mass_1 = total_mass / (1 + mass_ratio)
    mass_2 = mass_1 * mass_ratio
    return mass_1, mass_2


def bilby_spins_to_lalsim(
    theta_jn: BatchTensor,
    phi_jl: BatchTensor,
    tilt_1: BatchTensor,
    tilt_2: BatchTensor,
    phi_12: BatchTensor,
    a_1: BatchTensor,
    a_2: BatchTensor,
    mass_1: BatchTensor,
    mass_2: BatchTensor,
    f_ref: float,
    phi_ref: BatchTensor,
):
    """
    Converts between bilby spin and lalsimulation spin conventions.

    See https://github.com/bilby-dev/bilby/blob/cccdf891e82d46319e69dbfdf48c4970b4e9a727/bilby/gw/conversion.py#L105 # noqa
    and https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiral.c#L3594 # noqa

    Args:
        theta_jn: BatchTensor,
        phi_jl: BatchTensor,
        tilt_1: BatchTensor,
        tilt_2: BatchTensor,
        phi_12: BatchTensor,
        a_1: BatchTensor,
        a_2: BatchTensor,
        mass_1: BatchTensor,
        mass_2: BatchTensor,
        f_ref: float,
        phi_ref: BatchTensor,
    """

    # check if f_ref is valid
    if f_ref <= 0.0:
        raise ValueError(
            "f_ref <= 0 is invalid. "
            "Please pass in the starting GW frequency instead."
        )

    # starting frame: LNhat is along the z-axis and the unit
    # spin vectors are defined from the angles relative to LNhat.
    # Note that we put s1hat in the x-z plane, and phi12
    # sets the azimuthal angle of s2hat measured from the x-axis.
    lnh_x = 0
    lnh_y = 0
    lnh_z = 1
    # Spins are given wrt to L,
    # but still we cannot fill the spin as we do not know
    # what will be the relative orientation of L and N.
    # Note that these spin components are NOT wrt to binary
    # separation vector, but wrt to binary separation vector
    # at phiref=0.

    s1hatx = torch.sin(tilt_1) * torch.cos(phi_ref)
    s1haty = torch.sin(tilt_1) * torch.sin(phi_ref)
    s1hatz = torch.cos(tilt_1)
    s2hatx = torch.sin(tilt_2) * torch.cos(phi_12 + phi_ref)
    s2haty = torch.sin(tilt_2) * torch.sin(phi_12 + phi_ref)
    s2hatz = torch.cos(tilt_2)

    total_mass = mass_1 + mass_2

    eta = mass_1 * mass_2 / (mass_1 + mass_2) / (mass_1 + mass_2)

    # v parameter at reference point
    v0 = ((mass_1 + mass_2) * MTSUN_SI * PI * f_ref) ** (1 / 3)

    # Define S1, S2, J with proper magnitudes */

    l_mag = XLALSimInspiralLN(total_mass, eta, v0) * (
        1.0 + v0 * v0 * XLALSimInspiralL_2PN(eta)
    )
    s1x = mass_1 * mass_1 * a_1 * s1hatx
    s1y = mass_1 * mass_1 * a_1 * s1haty
    s1z = mass_1 * mass_1 * a_1 * s1hatz
    s2x = mass_2 * mass_2 * a_2 * s2hatx
    s2y = mass_2 * mass_2 * a_2 * s2haty
    s2z = mass_2 * mass_2 * a_2 * s2hatz
    Jx = s1x + s2x
    Jy = s1y + s2y
    Jz = l_mag + s1z + s2z

    # Normalize J to Jhat, find its angles in starting frame */
    Jnorm = torch.sqrt(Jx * Jx + Jy * Jy + Jz * Jz)
    Jhatx = Jx / Jnorm
    Jhaty = Jy / Jnorm
    Jhatz = Jz / Jnorm
    theta0 = torch.acos(Jhatz)
    phi0 = torch.atan2(Jhaty, Jhatx)

    # Rotation 1: Rotate about z-axis by -phi0 to put Jhat in x-z plane
    s1hatx, s1haty, s1hatz = rotate_z(-phi0, s1hatx, s1haty, s1hatz)
    s2hatx, s2haty, s2hatz = rotate_z(-phi0, s2hatx, s2haty, s2hatz)

    # Rotation 2: Rotate about new y-axis by -theta0
    # to put Jhat along z-axis

    lnh_x, lnh_y, lnh_z = rotate_y(-theta0, lnh_x, lnh_y, lnh_z)
    s1hatx, s1haty, s1hatz = rotate_y(-theta0, s1hatx, s1haty, s1hatz)
    s2hatx, s2haty, s2hatz = rotate_y(-theta0, s2hatx, s2haty, s2hatz)

    # Rotation 3: Rotate about new z-axis by phiJL to put L at desired
    # azimuth about J. Note that is currently in x-z plane towards -x
    # (i.e. azimuth=pi). Hence we rotate about z by phiJL - LAL_PI
    lnh_x, lnh_y, lnh_z = rotate_z(phi_jl - PI, lnh_x, lnh_y, lnh_z)
    s1hatx, s1haty, s1hatz = rotate_z(phi_jl - PI, s1hatx, s1haty, s1hatz)
    s2hatx, s2haty, s2hatz = rotate_z(phi_jl - PI, s2hatx, s2haty, s2hatz)

    # The cosinus of the angle between L and N is the scalar
    # product of the two vectors.
    # We do not need to perform additional rotation to compute it.
    Nx = 0.0
    Ny = torch.sin(theta_jn)
    Nz = torch.cos(theta_jn)
    incl = torch.acos(Nx * lnh_x + Ny * lnh_y + Nz * lnh_z)

    # Rotation 4-5: Now J is along z and N in y-z plane, inclined from J
    # by thetaJN and with >ve component along y.
    # Now we bring L into the z axis to get spin components.
    thetalj = torch.acos(lnh_z)
    phil = torch.atan2(lnh_y, lnh_x)

    s1hatx, s1haty, s1hatz = rotate_z(-phil, s1hatx, s1haty, s1hatz)
    s2hatx, s2haty, s2hatz = rotate_z(-phil, s2hatx, s2haty, s2hatz)
    Nx, Ny, Nz = rotate_z(-phil, Nx, Ny, Nz)

    s1hatx, s1haty, s1hatz = rotate_y(-thetalj, s1hatx, s1haty, s1hatz)
    s2hatx, s2haty, s2hatz = rotate_y(-thetalj, s2hatx, s2haty, s2hatz)
    Nx, Ny, Nz = rotate_y(-thetalj, Nx, Ny, Nz)

    # Rotation 6: Now L is along z and we have to bring N
    # in the y-z plane with >ve y components.

    phiN = torch.atan2(Ny, Nx)
    # Note the extra -phiRef here:
    # output spins must be given wrt to two body separations
    # which are rigidly rotated with spins
    s1hatx, s1haty, s1hatz = rotate_z(
        PI / 2.0 - phiN - phi_ref, s1hatx, s1haty, s1hatz
    )
    s2hatx, s2haty, s2hatz = rotate_z(
        PI / 2.0 - phiN - phi_ref, s2hatx, s2haty, s2hatz
    )

    s1x = s1hatx * a_1
    s1y = s1haty * a_1
    s1z = s1hatz * a_1
    s2x = s2hatx * a_2
    s2y = s2haty * a_2
    s2z = s2hatz * a_2

    return incl, s1x, s1y, s1z, s2x, s2y, s2z



import itertools
from typing import Optional, Tuple, Union

import torch
from jaxtyping import Float
from torch import Tensor

from ..utils.slicing import slice_kernels


class InMemoryDataset(torch.utils.data.IterableDataset):
    """Dataset for iterating through in-memory multi-channel timeseries

    Dataset for arrays of timeseries data which can be stored
    in-memory all at once. Iterates through the data by sampling
    fixed-length windows from all channels. The precise mechanism
    for this iteration is determined by combinations of the keyword
    arguments. See their descriptions for details.

    Args:
        X:
            Timeseries data to be iterated through. Should have
            shape `(num_channels, length * sample_rate)`. Windows
            will be sampled from the time (1st) dimension for all
            channels along the channel (0th) dimension.
        kernel_size:
            The length of the windows to sample from `X` in units
            of samples.
        y:
            Target timeseries to be iterated through. If specified,
            should be a single channel and have shape
            `(length * sample_rate,)`. If left as `None`, only windows
            sampled from `X` will be returned during iteration.
            Otherwise, windows sampled from both arrays will be
            returned. Note that if sampling is performed non-coincidentally,
            there's no sensible way to align windows sampled from this
            array with the windows sampled from `X`, so this combination
            of arguments is not permitted.
        batch_size:
            Maximum number of windows to return at each iteration. Will
            be the length of the 0th dimension of the returned array(s).
            If `batches_per_epoch` is specified, this will be the length
            of _every_ array returned during iteration. Otherwise, it's
            possible that the last array will be shorter due to the number
            of windows in the timeseries being a non-integer multiple of
            `batch_size`.
        stride:
            The resolution at which windows will be sampled from the
            specified timeseries, in units of samples. E.g. if
            `stride=2`, the first sample of each window can only be
            from an index of `X` which is a multiple of 2. Obviously,
            this reduces the number of windows which can be iterated
            through by a factor of `stride`.
        batches_per_epoch:
            Number of batches of window to produce during iteration
            before raising a `StopIteration`. Must be specified if
            performing non-coincident sampling. Otherwise, if left
            as `None`, windows will be sampled until the entire
            timeseries has been exhausted. Note that
            `batch_size * batches_per_epoch` must be be small
            enough to be able to be fulfilled by the number of
            windows in the timeseries, otherise a `ValueError`
            will be raised.
        coincident:
            Whether to sample windows from the channels of `X`
            using the same indices or independently. Can't be
            `True` if `batches_per_epoch` is `None` or `y` is
            _not_ `None`.
        shuffle:
            Whether to sample windows from timeseries randomly
            or in order along the time axis. If `coincident=False`
            and `shuffle=False`, channels will be iterated through
            with the index along the last channel moving fastest.
        device:
            Which device to host the timeseries arrays on
    """

    def __init__(
        self,
        X: Float[Tensor, "channels time"],
        kernel_size: int,
        y: Optional[Float[Tensor, " time"]] = None,
        batch_size: int = 32,
        stride: int = 1,
        batches_per_epoch: Optional[int] = None,
        coincident: bool = True,
        shuffle: bool = True,
        device: str = "cpu",
    ) -> None:
        self.X = torch.Tensor(X).to(device)

        # make sure if we specified a target array that all other
        # other necessary conditions are met (it has the same
        # length as `X` and we're sampling coincidentally)
        if y is not None and y.shape[-1] != X.shape[-1]:
            raise ValueError(
                "Target timeseries must have same length as input"
            )
        elif y is not None and not coincident:
            raise ValueError("Can't sample target array non-coincidentally")
        elif y is not None:
            self.y = y.to(device)
        else:
            self.y = None

        if not coincident and batches_per_epoch is None:
            # TODO: do we want to allow this? There are strides
            # for which this might be pallatable, but deciding
            # where to draw the line feels more complicated than
            # just allowing the user to compute the appropriate
            # batches_per_epoch themselves if they so choose
            raise ValueError(
                "Must specify number of batches between validation "
                "steps for non-coincident sampling"
            )

        # make sure that we'll have enough kernels to be able
        # to generate the specified number of batches
        self.kernel_size = kernel_size
        self.stride = stride
        if batches_per_epoch is not None and self.num_kernels < (
            batch_size * batches_per_epoch
        ):
            raise ValueError(
                "Number of kernels {} in timeseries insufficient "
                "to generate {} batches of size {}".format(
                    self.num_kernels, batch_size, batches_per_epoch
                )
            )

        self.batch_size = batch_size
        self.batches_per_epoch = batches_per_epoch
        self.shuffle = shuffle
        self.coincident = coincident

    @property
    def num_kernels(self) -> int:
        """
        The number of windows contained in the timeseries if we
        sample at the specified stride.
        """
        return (self.X.shape[-1] - self.kernel_size) // self.stride + 1

    def __len__(self) -> int:
        """
        The number of _batches_ contained in the timeseries
        """
        if self.batches_per_epoch is not None:
            return self.batches_per_epoch

        if self.coincident:
            return (self.num_kernels - 1) // self.batch_size + 1
        else:
            # TODO: this won't ever be triggered currently,
            # but leaving it in in case we ever choose to
            # support it
            num_kernels = self.num_kernels ** len(self.X)
            return (num_kernels - 1) // self.batch_size + 1

    def init_indices(self):
        """
        Initialize arrays of indices we'll use to slice
        through X and y at iteration time. This helps by
        taking care of building in any randomness upfront.
        """

        # establish how many kernels we'll actually be iterating through
        if self.batches_per_epoch is not None:
            num_kernels = self.batch_size * self.batches_per_epoch
        else:
            num_kernels = self.num_kernels

        # make sure the indices live on the same device as X (and y)
        device = self.X.device

        if not self.coincident and self.shuffle:
            # sampling non-coincidentally and randomly,
            # so generate sample indices for each channel
            # independently. In principle this means you
            # could sample the same set of kernels twice,
            # but for sensible numbers of kernels and channels
            # this likelihood is very low.
            idx = torch.randint(
                self.num_kernels,
                size=(num_kernels, len(self.X)),
                device=device,
            )
        elif not self.coincident:
            # sampling non-coincidentally but deterministically.
            # Use a little zip magic to only iterate through the
            # indices we'll need rather than having to generate
            # everything.
            idx = [range(self.num_kernels) for _ in range(len(self.X))]
            idx = zip(range(num_kernels), itertools.product(*idx))
            idx = torch.stack([torch.Tensor(i[1]) for i in idx])
            idx = idx.type(torch.int64).to(device)
        elif self.shuffle:
            # sampling randomly but coincidentally, so we
            # just need one set of indices and we can use
            # randperm to make sure we don't repeat
            idx = torch.randperm(self.num_kernels, device=device)
            idx = idx[:num_kernels]
        else:
            # the simplest case: deteriminstic and coincident
            idx = torch.arange(num_kernels, device=device)

        return idx

    def __iter__(
        self,
    ) -> Union[
        Float[Tensor, "batch channel time"],
        Tuple[Float[Tensor, "batch channel time"], Float[Tensor, " batch"]],
    ]:

        indices = self.init_indices()
        for i in range(len(self)):
            # slice the array of _indices_ we'll be using to
            # slice our timeseries, and scale them by the stride
            slc = slice(i * self.batch_size, (i + 1) * self.batch_size)
            idx = indices[slc] * self.stride

            # slice our timeseries
            X = slice_kernels(self.X, idx, self.kernel_size)
            if self.y is not None:
                y = slice_kernels(self.y, idx, self.kernel_size)
                yield X, y
            else:
                yield X



from typing import Optional, Union

import torch

from .. import spectral
from ..types import (
    FrequencySeries1d,
    FrequencySeries1to3d,
    TimeSeries1d,
    TimeSeries3d,
)
from .transform import FittableSpectralTransform


class Whiten(torch.nn.Module):
    """
    Normalize the frequency content of timeseries
    data by a provided power spectral density, such
    that if the timeseries are sampled from the same
    distribution as the PSD the normalized power will
    be approximately unity across all frequency bins.
    The whitened timeseries will then also have
    0 mean and unit variance.

    In order to avoid edge effects due to filter settle-in,
    the provided PSDs will have their spectrum truncated
    such that their impulse response time in the time
    domain is `fduration` seconds, and `fduration / 2`
    seconds worth of data will be removed from each
    edge of the whitened timeseries.

    For more information, see the documentation to
    `ml4gw.spectral.whiten`.

    Args:
        fduration:
            The length of the whitening filter's impulse
            response, in seconds. `fduration / 2` seconds
            worth of data will be cropped from the edges
            of the whitened timeseries.
        sample_rate:
            Rate at which timeseries data passed at call
            time is expected to be sampled
        highpass:
            Cutoff frequency to apply highpass filtering
            during whitening. If left as `None`, no highpass
            filtering will be performed.
        lowpass:
            Cutoff frequency to apply lowpass filtering
            during whitening. If left as `None`, no lowpass
            filtering will be performed.
    """

    def __init__(
        self,
        fduration: float,
        sample_rate: float,
        highpass: Optional[float] = None,
        lowpass: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.fduration = fduration
        self.sample_rate = sample_rate
        self.highpass = highpass
        self.lowpass = lowpass

        # register a window up front to signify our
        # fduration at inference time
        size = int(fduration * sample_rate)
        window = torch.hann_window(size, dtype=torch.float64)
        self.register_buffer("window", window)

    def forward(
        self, X: TimeSeries3d, psd: FrequencySeries1to3d
    ) -> TimeSeries3d:
        """
        Whiten a batch of multichannel timeseries by a
        background power spectral density.

        Args:
            X:
                Batch of multichannel timeseries to whiten.
                Should have the shape (B, C, N), where
                B is the batch size, C is the number of
                channels, and N is the number of seconds
                in the timeseries times `self.sample_rate`.
            psd:
                Power spectral density used to whiten the
                provided timeseries. Can be either 1D, 2D,
                or 3D, with the last dimension representing
                power at each frequency value. All other
                dimensions must match their corresponding
                value in `X`, starting from the right.
                (e.g. if `psd.ndim == 2`, `psd.size(1)` should
                be equal to `X.size(1)`. If `psd.ndim == 3`,
                `psd.size(1)` and `psd.size(0)` should be equal
                to `X.size(1)` and `X.size(0)`, respectively.)
                For more information about what these different
                shapes for `psd` represent, consult the documentation
                for `ml4gw.spectral.whiten`.
        Returns:
            Whitened timeseries, with `fduration * sample_rate / 2`
                samples cropped from each edge. Output shape will then
                be (B, C, N - `fduration * sample_rate`).
        """

        return spectral.whiten(
            X,
            psd,
            fduration=self.window,
            sample_rate=self.sample_rate,
            highpass=self.highpass,
            lowpass=self.lowpass,
        )


class FixedWhiten(FittableSpectralTransform):
    """
    Transform that whitens timeseries by a fixed
    power spectral density that's determined by
    calling the `.fit` method.

    Args:
        num_channels:
            Number of channels to whiten
        kernel_length:
            Expected length of tensors to whiten
            in seconds. Determines the number of
            frequency bins in the fit PSD.
        sample_rate:
            Rate at which timeseries will be sampled, in Hz
        dtype:
            Datatype with which background PSD will be stored
    """

    def __init__(
        self,
        num_channels: float,
        kernel_length: float,
        sample_rate: float,
        dtype: torch.dtype = torch.float64,
    ) -> None:
        super().__init__()
        self.num_channels = num_channels
        self.sample_rate = sample_rate
        self.kernel_length = kernel_length

        N = int(kernel_length * sample_rate)
        num_freqs = N // 2 + 1
        psd = torch.zeros((num_channels, num_freqs), dtype=dtype)
        self.register_buffer("psd", psd)

        # save this as a parameter since it's decided at fit time
        fduration = torch.zeros((1,))
        self.register_buffer("fduration", fduration)

    def fit(
        self,
        fduration: float,
        *background: Union[TimeSeries1d, FrequencySeries1d],
        fftlength: Optional[float] = None,
        highpass: Optional[float] = None,
        lowpass: Optional[float] = None,
        overlap: Optional[float] = None
    ) -> None:
        """
        Compute the PSD of channel-wise background to
        use to whiten timeseries at call time. PSDs will
        be resampled to have
        `self.kernel_length * self.sample_rate // 2 + 1`
        frequency bins.

        Args:
            fduration:
                Desired length of the impulse response
                of the whitening filter, in seconds.
                Fit PSDs will have their spectrum truncated
                to approximate this response time.
                A longer `fduration` will be able to
                handle narrower spikes in frequency, but
                at the expense of longer filter settle-in
                time. As such `fduration / 2` seconds of data
                will be removed from each edge of whitened
                timeseries.
            *background:
                1D arrays capturing the signal to be used to
                whiten each channel at call time. If `fftlength`
                is left as `None`, it will be assumed that these
                already represent frequency-domain data that will
                be possibly resampled and truncated to whiten
                timeseries at call time. Otherwise, it will be
                assumed that these represent time-domain data that
                will be converted to the frequency domain via
                Welch's method using the specified `fftlength`
                and `overlap`, with a Hann window used to window
                the FFT frames by default. Should have the same
                number of args as `self.num_channels`.
            fftlength:
                Length of frames used to convert time-domain
                data to the frequency-domain via Welch's method.
                If left as `None`, it will be assumed that the
                background arrays passed already represent frequency-
                domain data and don't require any conversion.
            highpass:
                Cutoff frequency, in Hz, used for highpass filtering
                with the fit whitening filter. This is achieved by
                setting the frequency response of the fit PSDs
                in the frequency bins below this value to 0.
                If left as `None`, the fit filter won't have any
                highpass filtering properties.
            lowpass:
                Cutoff frequency, in Hz, used for lowpass filtering
                with the fit whitening filter. This is achieved by
                setting the frequency response of the fit PSDs
                in the frequency bins above this value to 0.
                If left as `None`, the fit filter won't have any
                lowpass filtering properties.
            overlap:
                Overlap between FFT frames used to convert
                time-domain data to the frequency domain via
                Welch's method. If `fftlength` is `None`, this
                is ignored. Otherwise, if left as `None`, it will
                be set to half of `fftlength` by default.
        """
        if len(background) != self.num_channels:
            raise ValueError(
                "Expected to fit whitening transform on {} background "
                "timeseries, but was passed {}".format(
                    self.num_channels, len(background)
                )
            )

        num_freqs = self.psd.size(-1)
        psds = []
        for x in background:
            x = self.normalize_psd(
                x, self.sample_rate, num_freqs, fftlength, overlap
            )
            x = x.view(1, 1, -1)

            psd = spectral.truncate_inverse_power_spectrum(
                x, fduration, self.sample_rate, highpass, lowpass
            )
            psds.append(psd[0, 0])
        psd = torch.stack(psds)

        fduration = torch.Tensor([fduration])
        self.build(psd=psd, fduration=fduration)

    def forward(self, X: TimeSeries3d) -> TimeSeries3d:
        """
        Whiten the input timeseries tensor using the
        PSD fit by the `.fit` method, which must be
        called _before_ the first call to `.forward`.
        """
        expected_dim = int(self.kernel_length * self.sample_rate)
        if X.size(-1) != expected_dim:
            raise ValueError(
                "Whitening transform expected a kernel length "
                "of {}s, but was passed data of length {}s".format(
                    self.kernel_length, X.size(-1) / self.sample_rate
                )
            )

        pad = int(self.fduration.item() * self.sample_rate / 2)
        return spectral.normalize_by_psd(X, self.psd, self.sample_rate, pad)



import math
import warnings
from typing import List, Tuple

import torch
import torch.nn.functional as F
from jaxtyping import Float, Int
from torch import Tensor

from ..types import FrequencySeries1to3d, TimeSeries1to3d, TimeSeries3d
from .spline_interpolation import SplineInterpolate

"""
All based on https://github.com/gwpy/gwpy/blob/v3.0.8/gwpy/signal/qtransform.py
The methods, names, and descriptions come almost entirely from GWpy.
This code allows the Q-transform to be performed on batches of multi-channel
input on GPU.
"""


class QTile(torch.nn.Module):
    """
    Compute the row of Q-tiles for a single Q value and a single
    frequency for a batch of multi-channel frequency series data.
    Should really be called `QRow`, but I want to match GWpy.
    Input data should have three dimensions or fewer.
    If fewer, dimensions will be added until the input is
    three-dimensional.

    Args:
        q:
            The Q value to use in computing the Q tile
        frequency:
            The frequency for which to compute the Q tile in Hz
        duration:
            The length of time in seconds that the input frequency
            series represents
        sample_rate:
            The sample rate of the original time series in Hz
        mismatch:
            The maximum fractional mismatch between neighboring tiles

    """

    def __init__(
        self,
        q: float,
        frequency: float,
        duration: float,
        sample_rate: float,
        mismatch: float,
    ) -> None:
        super().__init__()
        self.mismatch = mismatch
        self.q = q
        self.deltam = torch.tensor(2 * (self.mismatch / 3.0) ** (1 / 2.0))
        self.qprime = self.q / 11 ** (1 / 2.0)
        self.frequency = frequency
        self.duration = duration
        self.sample_rate = sample_rate

        self.windowsize = (
            2 * int(self.frequency / self.qprime * self.duration) + 1
        )
        pad = self.ntiles() - self.windowsize
        padding = torch.Tensor((int((pad - 1) / 2.0), int((pad + 1) / 2.0)))
        self.register_buffer("padding", padding)
        self.register_buffer("indices", self.get_data_indices())
        self.register_buffer("window", self.get_window())

    def ntiles(self) -> int:
        """
        Number of tiles in this frequency row
        """
        tcum_mismatch = self.duration * 2 * torch.pi * self.frequency / self.q
        return int(2 ** torch.ceil(torch.log2(tcum_mismatch / self.deltam)))

    def _get_indices(self) -> Int[Tensor, " windowsize"]:
        half = int((self.windowsize - 1) / 2)
        return torch.arange(-half, half + 1)

    def get_window(self) -> Float[Tensor, " windowsize"]:
        """
        Generate the bi-square window for this row
        """
        wfrequencies = self._get_indices() / self.duration
        xfrequencies = wfrequencies * self.qprime / self.frequency
        norm = (
            self.ntiles()
            / (self.duration * self.sample_rate)
            * (315 * self.qprime / (128 * self.frequency)) ** (1 / 2.0)
        )
        return torch.Tensor((1 - xfrequencies**2) ** 2 * norm)

    def get_data_indices(self) -> Int[Tensor, " windowsize"]:
        """
        Get the index array of relevant frequencies for this row
        """
        return torch.round(
            self._get_indices() + 1 + self.frequency * self.duration,
        ).type(torch.long)

    def forward(
        self,
        fseries: FrequencySeries1to3d,
        norm: str = "median",
    ) -> TimeSeries1to3d:
        """
        Compute the transform for this row

        Args:
            fseries:
                Frequency series of data. Should correspond to data with
                the duration and sample rate used to initialize this object.
                Expected input shape is `(B, C, F)`, where F is the number
                of samples, C is the number of channels, and B is the number
                of batches. If less than three-dimensional, axes will be
                added.
            norm:
                The method of normalization. Options are "median", "mean", or
                `None`.

        Returns:
            The row of Q-tiles for the given Q and frequency. Output is
            three-dimensional: `(B, C, T)`
        """
        if len(fseries.shape) > 3:
            raise ValueError("Input data has more than 3 dimensions")

        while len(fseries.shape) < 3:
            fseries = fseries[None]

        windowed = fseries[..., self.indices] * self.window
        left, right = self.padding
        padded = F.pad(windowed, (int(left), int(right)), mode="constant")
        wenergy = torch.fft.ifftshift(padded, dim=-1)

        tdenergy = torch.fft.ifft(wenergy)
        energy = tdenergy.real**2.0 + tdenergy.imag**2.0
        if norm:
            norm = norm.lower() if isinstance(norm, str) else norm
            if norm == "median":
                medians = torch.quantile(energy, q=0.5, dim=-1, keepdim=True)
                energy /= medians
            elif norm == "mean":
                means = torch.mean(energy, dim=-1, keepdim=True)
                energy /= means
            else:
                raise ValueError("Invalid normalisation %r" % norm)
            energy = energy.type(torch.float32)
        return energy


class SingleQTransform(torch.nn.Module):
    """
    Compute the Q-transform for a single Q value for a batch of
    multi-channel time series data. Input data should have
    three dimensions or fewer.

    Args:
        duration:
            Length of the time series data in seconds
        sample_rate:
            Sample rate of the data in Hz
        spectrogram_shape:
            The shape of the interpolated spectrogram, specified as
            `(num_f_bins, num_t_bins)`. Because the
            frequency spacing of the Q-tiles is in log-space, the frequency
            interpolation is log-spaced as well.
        q:
            The Q value to use for the Q transform
        frange:
            The lower and upper frequency limit to consider for
            the transform. If unspecified, default values will
            be chosen based on q, sample_rate, and duration
        mismatch:
            The maximum fractional mismatch between neighboring tiles
        interpolation_method:
            The method by which to interpolate each `QTile` to the specified
            number of time and frequency bins. The acceptable values are
            "bilinear", "bicubic", and "spline". The "bilinear" and "bicubic"
            options will use PyTorch's built-in interpolation modes, while
            "spline" will use the custom Torch-based implementation in
            `ml4gw`, as PyTorch does not have spline-based intertpolation.
            The "spline" mode is most similar to the results of GWpy's
            Q-transform, which uses `scipy` to do spline interpolation.
            However, it is also the slowest and most memory intensive due to
            the matrix equation solving steps. Therefore, the default method
            is "bicubic" as it produces the most similar results while
            optimizing for computing performance.
    """

    def __init__(
        self,
        duration: float,
        sample_rate: float,
        spectrogram_shape: Tuple[int, int],
        q: float = 12,
        frange: List[float] = [0, torch.inf],
        mismatch: float = 0.2,
        interpolation_method: str = "bicubic",
    ) -> None:
        super().__init__()
        self.q = q
        self.spectrogram_shape = spectrogram_shape
        self.frange = frange
        self.duration = duration
        self.mismatch = mismatch

        # If q is too large, the minimum of the frange computed
        # below will be larger than the maximum
        max_q = torch.pi * duration * sample_rate / 50 - 11 ** (0.5)
        if q >= max_q:
            raise ValueError(
                "The given q value is too large for the given duration and "
                f"sample rate. The maximum allowable value is {max_q}"
            )

        if interpolation_method not in ["bilinear", "bicubic", "spline"]:
            raise ValueError(
                "Interpolation method must be either 'bilinear', 'bicubic', "
                f"or 'spline'; got {interpolation_method}"
            )
        self.interpolation_method = interpolation_method

        qprime = self.q / 11 ** (1 / 2.0)
        if self.frange[0] <= 0:  # set non-zero lower frequency
            self.frange[0] = 50 * self.q / (2 * torch.pi * duration)
        if math.isinf(self.frange[1]):  # set non-infinite upper frequency
            self.frange[1] = sample_rate / 2 / (1 + 1 / qprime)

        self.freqs = self.get_freqs()
        self.qtile_transforms = torch.nn.ModuleList(
            [
                QTile(
                    q=self.q,
                    frequency=freq,
                    duration=self.duration,
                    sample_rate=sample_rate,
                    mismatch=self.mismatch,
                )
                for freq in self.freqs
            ]
        )
        self.qtiles = None

        if self.interpolation_method == "spline":
            self._set_up_spline_interp()

    def _set_up_spline_interp(self):
        ntiles = [qtile.ntiles() for qtile in self.qtile_transforms]
        # For efficiency, we'll stack all qtiles of the same length before
        # interpolating, so we need to figure out which those are
        unique_ntiles = sorted(list(set(ntiles)))
        idx = torch.arange(len(ntiles))
        self.stack_idx = [idx[Tensor(ntiles) == n] for n in unique_ntiles]

        t_out = torch.arange(
            0, self.duration, self.duration / self.spectrogram_shape[1]
        )
        self.qtile_interpolators = torch.nn.ModuleList(
            [
                SplineInterpolate(
                    kx=3,
                    x_in=torch.arange(0, self.duration, self.duration / tiles),
                    y_in=torch.arange(len(idx)),
                    x_out=t_out,
                    y_out=torch.arange(len(idx)),
                )
                for tiles, idx in zip(unique_ntiles, self.stack_idx)
            ]
        )

        t_in = t_out
        f_in = self.freqs
        f_out = torch.logspace(
            math.log10(self.frange[0]),
            math.log10(self.frange[-1]),
            self.spectrogram_shape[0],
        )

        self.interpolator = SplineInterpolate(
            kx=3,
            ky=3,
            x_in=t_in,
            y_in=f_in,
            x_out=t_out,
            y_out=f_out,
        )

    def get_freqs(self) -> Float[Tensor, " nfreq"]:
        """
        Calculate the frequencies that will be used in this transform.
        For each frequency, a `QTile` is created.
        """
        minf, maxf = self.frange
        fcum_mismatch = (
            math.log(maxf / minf) * (2 + self.q**2) ** (1 / 2.0) / 2.0
        )
        deltam = 2 * (self.mismatch / 3.0) ** (1 / 2.0)
        nfreq = int(max(1, math.ceil(fcum_mismatch / deltam)))
        fstep = fcum_mismatch / nfreq
        fstepmin = 1 / self.duration

        freq_base = math.exp(2 / ((2 + self.q**2) ** (1 / 2.0)) * fstep)
        freqs = torch.Tensor([freq_base ** (i + 0.5) for i in range(nfreq)])
        # Cast freqs to float64 to avoid off-by-ones from rounding
        freqs = (minf * freqs.double() // fstepmin) * fstepmin
        return torch.unique(freqs)

    def get_max_energy(
        self, fsearch_range: List[float] = None, dimension: str = "both"
    ):
        """
        Gets the maximum energy value among the QTiles. The maximum can
        be computed across all batches and channels, across all channels,
        across all batches, or individually for each channel/batch
        combination. This could be useful for allowing the use of different
        Q values for different channels and batches, but the slicing would
        be slow, so this isn't used yet.

        Optionally, a pair of frequency values can be specified for
        `fsearch_range` to restrict the frequencies in which the maximum
        energy value is sought.
        """
        allowed_dimensions = ["both", "neither", "channel", "batch"]
        if dimension not in allowed_dimensions:
            raise ValueError(f"Dimension must be one of {allowed_dimensions}")

        if self.qtiles is None:
            raise RuntimeError(
                "Q-tiles must first be computed with .compute_qtiles()"
            )

        if fsearch_range is not None:
            start = min(torch.argwhere(self.freqs > fsearch_range[0]))
            stop = min(torch.argwhere(self.freqs > fsearch_range[1]))
            qtiles = self.qtiles[start:stop]
        else:
            qtiles = self.qtiles

        if dimension == "both":
            return max([torch.max(qtile) for qtile in qtiles])

        max_across_t = [torch.max(qtile, dim=-1).values for qtile in qtiles]
        max_across_t = torch.stack(max_across_t, dim=-1)
        max_across_ft = torch.max(max_across_t, dim=-1).values

        if dimension == "neither":
            return max_across_ft
        if dimension == "channel":
            return torch.max(max_across_ft, dim=-2).values
        if dimension == "batch":
            return torch.max(max_across_ft, dim=-1).values

    def compute_qtiles(
        self,
        X: TimeSeries1to3d,
        norm: str = "median",
    ) -> None:
        """
        Take the FFT of the input timeseries and calculate the transform
        for each `QTile`
        """
        # Computing the FFT with the same normalization and scaling as GWpy
        X = torch.fft.rfft(X, norm="forward")
        X[..., 1:] *= 2
        self.qtiles = [qtile(X, norm) for qtile in self.qtile_transforms]

    def interpolate(self) -> TimeSeries3d:
        if self.qtiles is None:
            raise RuntimeError(
                "Q-tiles must first be computed with .compute_qtiles()"
            )
        if self.interpolation_method == "spline":
            qtiles = [
                torch.stack([self.qtiles[i] for i in idx], dim=-2)
                for idx in self.stack_idx
            ]
            time_interped = torch.cat(
                [
                    interpolator(qtile)
                    for qtile, interpolator in zip(
                        qtiles, self.qtile_interpolators
                    )
                ],
                dim=-2,
            )
            return self.interpolator(time_interped)
        num_f_bins, num_t_bins = self.spectrogram_shape
        resampled = [
            F.interpolate(
                qtile[None],
                (qtile.shape[-2], num_t_bins),
                mode=self.interpolation_method,
            )
            for qtile in self.qtiles
        ]
        resampled = torch.stack(resampled, dim=-2)
        resampled = F.interpolate(
            resampled[0],
            (num_f_bins, num_t_bins),
            mode=self.interpolation_method,
        )
        return torch.squeeze(resampled)

    def forward(
        self,
        X: TimeSeries1to3d,
        norm: str = "median",
    ):
        """
        Compute the Q-tiles and interpolate

        Args:
            X:
                Time series of data. Should have the duration and sample rate
                used to initialize this object. Expected input shape is
                `(B, C, T)`, where T is the number of samples, C is the number
                of channels, and B is the number of batches. If less than
                three-dimensional, axes will be added during Q-tile
                computation.
            norm:
                The method of normalization used by each QTile

        Returns:
            The interpolated Q-transform for the batch of data. Output will
            have one more dimension than the input
        """

        self.compute_qtiles(X, norm)
        return self.interpolate()


class QScan(torch.nn.Module):
    """
    Calculate the Q-transform of a batch of multi-channel
    time series data for a range of Q values and return
    the interpolated Q-transform with the highest energy.

    Args:
        duration:
            Length of the time series data in seconds
        sample_rate:
            Sample rate of the data in Hz
        spectrogram_shape:
            The shape of the interpolated spectrogram, specified as
            `(num_f_bins, num_t_bins)`. Because the
            frequency spacing of the Q-tiles is in log-space, the frequency
            interpolation is log-spaced as well.
        qrange:
            The lower and upper values of Q to consider. The
            actual values of Q used for the transforms are
            determined by the `get_qs` method
        frange:
            The lower and upper frequency limit to consider for
            the transform. If unspecified, default values will
            be chosen based on q, sample_rate, and duration
        mismatch:
            The maximum fractional mismatch between neighboring tiles
    """

    def __init__(
        self,
        duration: float,
        sample_rate: float,
        spectrogram_shape: Tuple[int, int],
        qrange: List[float] = [4, 64],
        frange: List[float] = [0, torch.inf],
        interpolation_method="bicubic",
        mismatch: float = 0.2,
    ) -> None:
        super().__init__()
        self.qrange = qrange
        self.mismatch = mismatch
        self.frange = frange
        self.spectrogram_shape = spectrogram_shape
        max_q = torch.pi * duration * sample_rate / 50 - 11 ** (0.5)
        self.qs = self.get_qs()
        if self.qs[-1] >= max_q:
            warnings.warn(
                "Some Q values exceed the maximum allowable Q value of "
                f"{max_q}. The list of Q values to be tested in this "
                "scan will be truncated to avoid those values."
            )

        # Deliberately doing something different from GWpy here.
        # Their final frange is the intersection of the frange
        # from each q. This implementation uses the frange of
        # the chosen q.
        self.q_transforms = torch.nn.ModuleList(
            [
                SingleQTransform(
                    duration=duration,
                    sample_rate=sample_rate,
                    spectrogram_shape=spectrogram_shape,
                    q=q,
                    frange=self.frange.copy(),
                    interpolation_method=interpolation_method,
                    mismatch=self.mismatch,
                )
                for q in self.qs
                if q < max_q
            ]
        )

    def get_qs(self) -> List[float]:
        """
        Determine the values of Q to try for the set of Q-transforms
        """
        deltam = 2 * (self.mismatch / 3.0) ** (1 / 2.0)
        cumum = math.log(self.qrange[1] / self.qrange[0]) / 2 ** (1 / 2.0)
        nplanes = int(max(math.ceil(cumum / deltam), 1))
        dq = cumum / nplanes
        qs = [
            self.qrange[0] * math.exp(2 ** (1 / 2.0) * dq * (i + 0.5))
            for i in range(nplanes)
        ]

        return qs

    def forward(
        self,
        X: TimeSeries1to3d,
        fsearch_range: List[float] = None,
        norm: str = "median",
    ):
        """
        Compute the set of QTiles for each Q transform and determine which
        has the highest energy value. Interpolate and return the
        corresponding set of tiles.

        Args:
            X:
                Time series of data. Should have the duration and sample rate
                used to initialize this object. Expected input shape is
                `(B, C, T)`, where T is the number of samples, C is the number
                of channels, and B is the number of batches. If less than
                three-dimensional, axes will be added during Q-tile
                computation.
            fsearch_range:
                The lower and upper frequency values within which to search
                for the maximum energy
            norm:
                The method of interpolation used by each QTile

        Returns:
            An interpolated Q-transform for the batch of data. Output will
            have one more dimension than the input
        """
        for transform in self.q_transforms:
            transform.compute_qtiles(X, norm)
        idx = torch.argmax(
            torch.Tensor(
                [
                    transform.get_max_energy(fsearch_range=fsearch_range)
                    for transform in self.q_transforms
                ]
            )
        )
        return self.q_transforms[idx].interpolate()



from typing import Optional, Sequence, Tuple

import torch
from jaxtyping import Float
from torch import Tensor

from ...utils.slicing import unfold_windows


class Snapshotter(torch.nn.Module):
    """
    Model for converting streaming state updates into
    a batch of overlapping snaphots of a multichannel
    timeseries. Can support multiple timeseries in a
    single state update via the `channels_per_snapshot`
    kwarg.

    Specifically, maps tensors of shape
    `(num_channels, batch_size * stride_size)` to a tensor
    of shape `(batch_size, num_channels, snapshot_size)`.
    If `channels_per_snapshot` is specified, it will return
    `len(channels_per_snapshot)` tensors of this shape,
    with the channel dimension replaced by the corresponding
    value of `channels_per_snapshot`. The last tensor returned
    at call time will be the current state that can be passed
    to the next `forward` call.

    Args:
        num_channels:
            Number of channels in the timeseries. If
            `channels_per_snapshot` is not `None`,
            this should be equal to `sum(channels_per_snapshot)`.
        snapshot_size:
            The size of the output snapshot windows in
            number of samples
        stride_size:
            The number of samples in between each output
            snapshot
        batch_size:
            The number of snapshots to produce at each
            update. The last dimension of the input
            tensor should have size `batch_size * stride_size`.
        channels_per_snapshot:
            How to split up the channels in the timeseries
            for different tensors. If left as `None`, all
            the channels will be returned in a single tensor.
            Otherwise, the channels will be split up into
            `len(channels_per_snapshot)` tensors, with each
            tensor's channel dimension being equal to the
            corresponding value in `channels_per_snapshot`.
            Therefore, if specified, these values should
            add up to `num_channels`.
    """

    def __init__(
        self,
        num_channels: int,
        snapshot_size: int,
        stride_size: int,
        batch_size: int,
        channels_per_snapshot: Optional[Sequence[int]] = None,
    ) -> None:
        super().__init__()
        if stride_size >= snapshot_size:
            raise ValueError(
                "Snapshotter can't accommodate stride {} "
                "which is greater than snapshot size {}".format(
                    stride_size, snapshot_size
                )
            )

        self.snapshot_size = snapshot_size
        self.stride_size = stride_size
        self.state_size = snapshot_size - stride_size
        self.batch_size = batch_size

        if channels_per_snapshot is not None:
            if sum(channels_per_snapshot) != num_channels:
                raise ValueError(
                    "Can't break {} channels into {}".format(
                        num_channels, channels_per_snapshot
                    )
                )
        self.channels_per_snapshot = channels_per_snapshot
        self.num_channels = num_channels

    def get_initial_state(self) -> Float[Tensor, "channel time"]:
        return torch.zeros((self.num_channels, self.state_size))

    def forward(
        self,
        update: Float[Tensor, "channel time1"],
        snapshot: Optional[Float[Tensor, "channel time2"]] = None,
    ) -> Tuple[Tensor, ...]:
        if snapshot is None:
            snapshot = self.get_initial_state()

        # append new data to the snapshot
        snapshot = torch.cat([snapshot, update], axis=-1)

        if self.batch_size > 1:
            snapshots = unfold_windows(
                snapshot, self.snapshot_size, self.stride_size
            )
        else:
            snapshots = snapshot[None]

        if self.channels_per_snapshot is not None:
            if snapshots.size(1) != self.num_channels:
                raise ValueError(
                    "Expected {} channels, found {}".format(
                        self.num_channels, snapshots.size(1)
                    )
                )
            snapshots = torch.split(
                snapshots, self.channels_per_snapshot, dim=1
            )
        else:
            snapshots = (snapshots,)

        # keep only the latest snapshot as our state
        snapshot = snapshot[:, -self.state_size :]
        return tuple(snapshots) + (snapshot,)



from collections.abc import Sequence
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from .skip_connection import SkipConnection


class Autoencoder(torch.nn.Module):
    """
    Base autoencoder class that defines some of the
    basic methods and functionality. Autoencoders are
    defined here as a set of sequential blocks that
    have an `encode` method, which acts on the input
    data to the autoencoder, and a `decode` method, which
    acts on the encoded vector generated by the `encode`
    method. `forward` just runs these steps one after the
    other. Although it isn't explicitly enforced, a good
    rule of thumb is that the ouput of a block's `decode`
    method should have the same shape as the _input_ of its
    `encode` method.

    Accepts a `skip_connection` argument that defines how to
    combine information from the input of one block's `encode`
    layer with the output to its `decode`layer. See `skip_connections.py`
    for more info about what these classes are expected to contain
    and how they operate.
    """

    def __init__(
        self, skip_connection: Optional[SkipConnection] = None
    ) -> None:
        super().__init__()
        self.skip_connection = skip_connection
        self.blocks = torch.nn.ModuleList()

    def encode(
        self, *X: Tensor, return_states: bool = False
    ) -> Union[Tensor, Tuple[Tensor, Sequence]]:
        states = []
        for block in self.blocks:
            if isinstance(X, tuple):
                X = block.encode(*X)
            else:
                X = block.encode(X)
            states.append(X)

        # don't need to return the last
        # state, since that's just equal
        # to the output of this layer
        if return_states:
            return X, states[:-1]
        return X

    def decode(self, *X, states: Optional[Sequence[Tensor]] = None) -> Tensor:
        if self.skip_connection is not None and states is None:
            raise ValueError(
                "Must pass intermediate states when autoencoder "
                "has a skip connection function specified"
            )
        elif states is not None:
            if len(states) != len(self.blocks) - 1:
                raise ValueError(
                    "Passed {} intermediate states, expected {}".format(
                        len(states), len(self.blocks) - 1
                    )
                )

            # Don't skip connect the output layer
            states = states[::-1] + [None]

        for i, block in enumerate(self.blocks[::-1]):
            if isinstance(X, tuple):
                X = block.decode(*X)
            else:
                X = block.decode(X)

            state = states[-i - 1]
            if state is not None:
                X = self.skip_connection(X, state)
        return X

    def forward(self, *X: Tensor) -> Tensor:
        return_states = self.skip_connection is not None
        X = self.encode(*X, return_states=return_states)
        if return_states:
            *X, states = X
        else:
            states = None

        if isinstance(X, Tensor):
            X = (X,)
        return self.decode(*X, states=states)



import warnings
from typing import Optional, Sequence, Union

import h5py
import numpy as np
import torch

from ..types import WaveformTensor


class ContiguousHdf5Warning(Warning):
    pass


class Hdf5TimeSeriesDataset(torch.utils.data.IterableDataset):
    """
    Iterable dataset that samples and loads windows of
    timeseries data uniformly from a set of HDF5 files.
    It is _strongly_ recommended that these files have been
    written using [chunked storage]
    (https://docs.h5py.org/en/stable/high/dataset.html#chunked-storage).
    This has shown to produce increases in read-time speeds
    of over an order of magnitude.

    Args:
        fnames:
            Paths to HDF5 files from which to sample data.
        channels:
            Datasets to read from the indicated files, which
            will be stacked along dim 1 of the generated batches
            during iteration.
        kernel_size:
            Size of the windows to read, in number of samples.
            This will be the size of the last dimension of the
            generated batches.
        batch_size:
            Number of windows to sample at each iteration.
        batches_per_epoch:
            Number of batches to generate during each call
            to `__iter__`.
        coincident:
            Whether windows for each channel in a given batch
            element should be sampled coincidentally, i.e.
            corresponding to the same time indices from the
            same files, or should be sampled independently.
            For the latter case, users can either specify
            `False`, which will sample filenames independently
            for each channel, or `"files"`, which will sample
            windows independently within a given file for each
            channel. The latter setting limits the amount of
            entropy in the effective dataset, but can provide
            over 2x improvement in total throughput.
        num_files_per_batch:
            The number of unique files from which to sample
            batch elements each epoch. If left as `None`,
            will use all available files. Useful when reading
            from many files is bottlenecking dataloading.


    """

    def __init__(
        self,
        fnames: Sequence[str],
        channels: Sequence[str],
        kernel_size: int,
        batch_size: int,
        batches_per_epoch: int,
        coincident: Union[bool, str],
        num_files_per_batch: Optional[int] = None,
    ) -> None:
        if not isinstance(coincident, bool) and coincident != "files":
            raise ValueError(
                "coincident must be either a boolean or 'files', "
                "got unrecognized value {}".format(coincident)
            )

        self.fnames = np.array(fnames)
        self.channels = channels
        self.num_channels = len(channels)
        self.kernel_size = kernel_size
        self.batch_size = batch_size
        self.batches_per_epoch = batches_per_epoch
        self.coincident = coincident
        self.num_files_per_batch = (
            len(fnames) if num_files_per_batch is None else num_files_per_batch
        )
        if self.num_files_per_batch > len(fnames):
            raise ValueError(
                f"Number of files per batch ({self.num_files_per_batch}) "
                f"cannot exceed number of files ({len(fnames)}) "
            )

        self.sizes = {}
        for fname in self.fnames:
            with h5py.File(fname, "r") as f:
                dset = f[channels[0]]
                if dset.chunks is None:
                    warnings.warn(
                        "File {} contains datasets that were generated "
                        "without using chunked storage. This can have "
                        "severe performance impacts at data loading time. "
                        "If you need faster loading, try re-generating "
                        "your dataset with chunked storage turned on.".format(
                            fname
                        ),
                        category=ContiguousHdf5Warning,
                    )

                self.sizes[fname] = len(dset)

        total = sum(self.sizes.values())
        self.probs = np.array([i / total for i in self.sizes.values()])

    def __len__(self) -> int:
        return self.batches_per_epoch

    def sample_fnames(self, size) -> np.ndarray:
        # first, randomly select `self.num_files_per_batch`
        # file indices based on their probabilities
        fname_indices = np.arange(len(self.fnames))
        fname_indices = np.random.choice(
            fname_indices,
            p=self.probs,
            size=(self.num_files_per_batch),
            replace=False,
        )
        # now renormalize the probabilities, and sample
        # the requested size from this subset of files
        probs = self.probs[fname_indices]
        probs /= probs.sum()
        return np.random.choice(
            self.fnames[fname_indices],
            p=probs,
            size=size,
            replace=True,
        )

    def sample_batch(self) -> WaveformTensor:
        """
        Sample a single batch of multichannel timeseries
        """

        # allocate memory up front
        x = np.zeros((self.batch_size, len(self.channels), self.kernel_size))

        # sample filenames, but only loop through each unique
        # filename once to avoid unnecessary I/O overhead
        if self.coincident is not False:
            size = (self.batch_size,)
        else:
            size = (self.batch_size, self.num_channels)
        fnames = self.sample_fnames(size)

        unique_fnames, inv, counts = np.unique(
            fnames, return_inverse=True, return_counts=True
        )
        for i, (fname, count) in enumerate(zip(unique_fnames, counts)):
            size = self.sizes[fname]
            max_idx = size - self.kernel_size

            # figure out which batch indices should be
            # sampled from the current filename
            indices = np.where(inv == i)[0]

            # when sampling coincidentally either fully
            # or at the file level, all channels will
            # correspond to the same file
            if self.coincident is not False:
                batch_indices = np.repeat(indices, self.num_channels)
                channel_indices = np.arange(self.num_channels)
                channel_indices = np.concatenate([channel_indices] * count)
            else:
                batch_indices = indices // self.num_channels
                channel_indices = indices % self.num_channels

            # if we're sampling fully coincidentally, each
            # channel will be the same in each file
            if self.coincident is True:
                idx = np.random.randint(max_idx, size=count)
                idx = np.repeat(idx, self.num_channels)
            else:
                # otherwise, every channel will be different
                # for the given file
                idx = np.random.randint(max_idx, size=len(batch_indices))

            # open the file and sample a different set of
            # kernels for each batch element it occupies
            with h5py.File(fname, "r") as f:
                for b, c, i in zip(batch_indices, channel_indices, idx):
                    x[b, c] = f[self.channels[c]][i : i + self.kernel_size]
        return torch.Tensor(x)

    def __iter__(self) -> WaveformTensor:
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            num_batches = self.batches_per_epoch
        else:
            num_batches, remainder = divmod(
                self.batches_per_epoch, worker_info.num_workers
            )
            if worker_info.id < remainder:
                num_batches += 1

        for _ in range(num_batches):
            yield self.sample_batch()



"""
    Based on the JAX implementation of IMRPhenomPv2 from
    https://github.com/tedwards2412/ripple/blob/main/src/ripplegw/waveforms/IMRPhenomPv2.py
"""

from typing import Dict, Optional, Tuple

import torch
from jaxtyping import Float
from torch import Tensor

from ...constants import MPC_SEC, MTSUN_SI, PI
from ...types import BatchTensor, FrequencySeries1d
from ..conversion import (
    chirp_mass_and_mass_ratio_to_components,
    rotate_y,
    rotate_z,
)
from .phenom_d import IMRPhenomD


class IMRPhenomPv2(IMRPhenomD):
    def __init__(self):
        super().__init__()

    def forward(
        self,
        fs: FrequencySeries1d,
        chirp_mass: BatchTensor,
        mass_ratio: BatchTensor,
        s1x: BatchTensor,
        s1y: BatchTensor,
        s1z: BatchTensor,
        s2x: BatchTensor,
        s2y: BatchTensor,
        s2z: BatchTensor,
        distance: BatchTensor,
        phic: BatchTensor,
        inclination: BatchTensor,
        f_ref: float,
        tc: Optional[BatchTensor] = None,
        **kwargs,
    ):
        """
        IMRPhenomPv2 waveform

        Args:
            fs :
                Frequency series in Hz.
            chirp_mass :
                Chirp mass in solar masses.
            mass_ratio :
                Mass ratio m1/m2.
            s1x :
                Spin component x of the first BH.
            s1y :
                Spin component y of the first BH.
            s1z :
                Spin component z of the first BH.
            s2x :
                Spin component x of the second BH.
            s2y :
                Spin component y of the second BH.
            s2z :
                Spin component z of the second BH.
            distance :
                Luminosity distance in Mpc.
            tc :
                Coalescence time.
            phic :
                Reference phase.
            inclination :
                Inclination angle.
            f_ref :
                Reference frequency in Hz.

        Returns:
            hc, hp: Tuple[torch.Tensor, torch.Tensor]
                Cross and plus polarizations

        Note: m1 must be larger than m2.
        """

        if tc is None:
            tc = torch.zeros_like(chirp_mass)

        m1, m2 = chirp_mass_and_mass_ratio_to_components(
            chirp_mass, mass_ratio
        )

        # # flip m1 m2. For some reason LAL uses this convention for PhenomPv2
        m1, m2 = m2, m1
        s1x, s2x = s2x, s1x
        s1y, s2y = s2y, s1y
        s1z, s2z = s2z, s1z
        (
            chi1_l,
            chi2_l,
            chip,
            thetaJN,
            alpha0,
            phi_aligned,
            zeta_polariz,
        ) = self.convert_spins(
            m1, m2, f_ref, phic, inclination, s1x, s1y, s1z, s2x, s2y, s2z
        )

        phic = 2 * phi_aligned
        q = m2 / m1  # q>=1
        M = m1 + m2
        chi_eff = (m1 * chi1_l + m2 * chi2_l) / M
        chil = (1.0 + q) / q * chi_eff
        eta = m1 * m2 / (M * M)
        eta2 = eta * eta
        Seta = torch.sqrt(1.0 - 4.0 * eta)
        chi = self.chiPN(Seta, eta, chi2_l, chi1_l)
        chi22 = chi2_l * chi2_l
        chi12 = chi1_l * chi1_l
        xi = -1.0 + chi
        m_sec = M * MTSUN_SI
        piM = PI * m_sec

        omega_ref = piM * f_ref
        logomega_ref = torch.log(omega_ref)
        omega_ref_cbrt = (piM * f_ref) ** (1 / 3)  # == v0
        omega_ref_cbrt2 = omega_ref_cbrt * omega_ref_cbrt

        angcoeffs = self.ComputeNNLOanglecoeffs(q, chil, chip)

        alphaNNLOoffset = (
            angcoeffs["alphacoeff1"] / omega_ref
            + angcoeffs["alphacoeff2"] / omega_ref_cbrt2
            + angcoeffs["alphacoeff3"] / omega_ref_cbrt
            + angcoeffs["alphacoeff4"] * logomega_ref
            + angcoeffs["alphacoeff5"] * omega_ref_cbrt
        )

        epsilonNNLOoffset = (
            angcoeffs["epsiloncoeff1"] / omega_ref
            + angcoeffs["epsiloncoeff2"] / omega_ref_cbrt2
            + angcoeffs["epsiloncoeff3"] / omega_ref_cbrt
            + angcoeffs["epsiloncoeff4"] * logomega_ref
            + angcoeffs["epsiloncoeff5"] * omega_ref_cbrt
        )

        Y2m2 = self.SpinWeightedY(thetaJN, 0, -2, 2, -2)
        Y2m1 = self.SpinWeightedY(thetaJN, 0, -2, 2, -1)
        Y20 = self.SpinWeightedY(thetaJN, 0, -2, 2, -0)
        Y21 = self.SpinWeightedY(thetaJN, 0, -2, 2, 1)
        Y22 = self.SpinWeightedY(thetaJN, 0, -2, 2, 2)
        Y2 = torch.stack((Y2m2, Y2m1, Y20, Y21, Y22))

        hPhenomDs, diffRDphase = self.PhenomPOneFrequency(
            fs,
            m2,
            m1,
            eta,
            eta2,
            Seta,
            chi2_l,
            chi1_l,
            chi12,
            chi22,
            chip,
            phic,
            M,
            xi,
            distance,
        )

        hp, hc = self.PhenomPCoreTwistUp(
            fs,
            hPhenomDs,
            eta,
            chi1_l,
            chi2_l,
            chip,
            M,
            angcoeffs,
            Y2,
            alphaNNLOoffset - alpha0,
            epsilonNNLOoffset,
        )
        t0 = (diffRDphase.unsqueeze(1)) / (2 * PI)
        phase_corr = torch.cos(2 * PI * fs * (t0)) - 1j * torch.sin(
            2 * PI * fs * (t0)
        )
        M_s = (m1 + m2) * MTSUN_SI
        phase_corr_tc = torch.exp(
            -1j * fs * M_s.unsqueeze(1) * tc.unsqueeze(1)
        )
        hp *= phase_corr * phase_corr_tc
        hc *= phase_corr * phase_corr_tc

        c2z = torch.cos(2 * zeta_polariz).unsqueeze(1)
        s2z = torch.sin(2 * zeta_polariz).unsqueeze(1)
        hplus = c2z * hp + s2z * hc
        hcross = c2z * hc - s2z * hp
        return hcross, hplus

    def PhenomPCoreTwistUp(
        self,
        fHz: FrequencySeries1d,
        hPhenom: BatchTensor,
        eta: BatchTensor,
        chi1_l: BatchTensor,
        chi2_l: BatchTensor,
        chip: BatchTensor,
        M: BatchTensor,
        angcoeffs: Dict[str, BatchTensor],
        Y2m: BatchTensor,
        alphaoffset: BatchTensor,
        epsilonoffset: BatchTensor,
    ) -> Tuple[BatchTensor, BatchTensor]:
        assert angcoeffs is not None
        assert Y2m is not None
        f = fHz * MTSUN_SI * M.unsqueeze(1)  # Frequency in geometric units
        q = (1.0 + torch.sqrt(1.0 - 4.0 * eta) - 2.0 * eta) / (2.0 * eta)
        m1 = 1.0 / (1.0 + q)  # Mass of the smaller BH for unit total mass M=1.
        m2 = q / (1.0 + q)  # Mass of the larger BH for unit total mass M=1.
        Sperp = chip * (
            m2 * m2
        )  # Dimensionfull spin component in the orbital plane.
        # S_perp = S_2_perp chi_eff = m1 * chi1_l + m2 * chi2_l
        # effective spin for M=1

        SL = chi1_l * m1 * m1 + chi2_l * m2 * m2  # Dimensionfull aligned spin.

        omega = PI * f
        logomega = torch.log(omega)
        omega_cbrt = (omega) ** (1 / 3)
        omega_cbrt2 = omega_cbrt * omega_cbrt
        alpha = (
            (
                angcoeffs["alphacoeff1"] / omega.mT
                + angcoeffs["alphacoeff2"] / omega_cbrt2.mT
                + angcoeffs["alphacoeff3"] / omega_cbrt.mT
                + angcoeffs["alphacoeff4"] * logomega.mT
                + angcoeffs["alphacoeff5"] * omega_cbrt.mT
            )
            - alphaoffset
        ).mT

        epsilon = (
            (
                angcoeffs["epsiloncoeff1"] / omega.mT
                + angcoeffs["epsiloncoeff2"] / omega_cbrt2.mT
                + angcoeffs["epsiloncoeff3"] / omega_cbrt.mT
                + angcoeffs["epsiloncoeff4"] * logomega.mT
                + angcoeffs["epsiloncoeff5"] * omega_cbrt.mT
            )
            - epsilonoffset
        ).mT

        cBetah, sBetah = self.WignerdCoefficients(
            omega_cbrt.mT, SL, eta, Sperp
        )

        cBetah2 = cBetah * cBetah
        cBetah3 = cBetah2 * cBetah
        cBetah4 = cBetah3 * cBetah
        sBetah2 = sBetah * sBetah
        sBetah3 = sBetah2 * sBetah
        sBetah4 = sBetah3 * sBetah

        hp_sum = 0
        hc_sum = 0

        cexp_i_alpha = torch.exp(1j * alpha)
        cexp_2i_alpha = cexp_i_alpha * cexp_i_alpha
        cexp_mi_alpha = 1.0 / cexp_i_alpha
        cexp_m2i_alpha = cexp_mi_alpha * cexp_mi_alpha
        T2m = (
            cexp_2i_alpha.mT * cBetah4.mT * Y2m[0]
            - cexp_i_alpha.mT * 2 * cBetah3.mT * sBetah.mT * Y2m[1]
            + 1
            * torch.sqrt(torch.tensor(6))
            * sBetah2.mT
            * cBetah2.mT
            * Y2m[2]
            - cexp_mi_alpha.mT * 2 * cBetah.mT * sBetah3.mT * Y2m[3]
            + cexp_m2i_alpha.mT * sBetah4.mT * Y2m[4]
        ).mT
        Tm2m = (
            cexp_m2i_alpha.mT * sBetah4.mT * torch.conj(Y2m[0])
            + cexp_mi_alpha.mT
            * 2
            * cBetah.mT
            * sBetah3.mT
            * torch.conj(Y2m[1])
            + 1
            * torch.sqrt(torch.tensor(6))
            * sBetah2.mT
            * cBetah2.mT
            * torch.conj(Y2m[2])
            + cexp_i_alpha.mT * 2 * cBetah3.mT * sBetah.mT * torch.conj(Y2m[3])
            + cexp_2i_alpha.mT * cBetah4.mT * torch.conj(Y2m[4])
        ).mT
        hp_sum = T2m + Tm2m
        hc_sum = 1j * (T2m - Tm2m)

        eps_phase_hP = torch.exp(-2j * epsilon) * hPhenom / 2.0

        hp = eps_phase_hP * hp_sum
        hc = eps_phase_hP * hc_sum

        return hp, hc

    def PhenomPOneFrequency(
        self,
        fs,
        m1,
        m2,
        eta,
        eta2,
        Seta,
        chi1,
        chi2,
        chi12,
        chi22,
        chip,
        phic,
        M,
        xi,
        distance,
    ):
        """
        m1, m2: in solar masses
        phic: Orbital phase at peak of the underlying non precessing model
        M: Total mass (Solar masses)
        """
        M_s = M * MTSUN_SI
        Mf = torch.outer(M_s, fs)
        fRD, fDM = self.phP_get_fRD_fdamp(m1, m2, chi1, chi2, chip)
        # pass M_s * ringdown and M_s * damping frequency to PhenomD functions
        MfRD, MfDM = M_s * fRD, M_s * fDM

        phase, _ = self.phenom_d_phase(
            Mf, m1, m2, eta, eta2, chi1, chi2, xi, MfRD, MfDM
        )
        phase = (phase.mT - (phic + PI / 4.0)).mT
        # why are they subtracting 2*phic?
        # https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimIMRPhenomP.c#L1316

        Amp = self.phenom_d_amp(
            Mf,
            m1,
            m2,
            eta,
            eta2,
            Seta,
            chi1,
            chi2,
            chi12,
            chi22,
            xi,
            distance,
            MfRD,
            MfDM,
        )[0]
        Amp0 = self.get_Amp0(Mf, eta)
        dist_s = distance * MPC_SEC
        Amp = ((Amp0 * Amp).mT * (M_s**2.0) / dist_s).mT

        hPhenom = Amp * (torch.exp(-1j * phase))

        # calculating derivative of phase with frequency following
        # https://git.ligo.org/lscsoft/lalsuite/-/blame/master/lalsimulation/lib/LALSimIMRPhenomP.c?page=2#L1057 # noqa: E501
        n_fixed = 1000
        x = torch.linspace(0.8, 1.2, n_fixed, device=fRD.device)
        fRDs = torch.outer(fRD, x)
        delta_fRds = (1.2 * fRD - 0.8 * fRD) / (n_fixed - 1)
        MfRDs = torch.zeros_like(fRDs)
        for i in range(fRD.shape[0]):
            MfRDs[i, :] = torch.outer(M_s, fRDs[i, :])[i, :]
        RD_phase = self.phenom_d_phase(
            MfRDs, m1, m2, eta, eta2, chi1, chi2, xi, MfRD, MfDM
        )[0]
        diff = torch.diff(RD_phase, axis=1)
        diffRDphase = (diff[:, 1:] + diff[:, :-1]) / (
            2 * delta_fRds.unsqueeze(1)
        )
        # reshape x to have same shape as diffRDphase
        x = x[1:-1].unsqueeze(0).expand(diffRDphase.shape)
        # interpolate at x = 1, as thats the same as f = fRD
        diffRDphase = -self.interpolate(
            torch.tensor([1], device=x.device), x, diffRDphase
        )
        return hPhenom, diffRDphase

    # Utility functions

    def interpolate(
        self,
        x: Float[Tensor, " new_series"],
        xp: Float[Tensor, " series"],
        fp: Float[Tensor, " series"],
    ) -> Float[Tensor, " new_series"]:
        """One-dimensional linear interpolation for monotonically
        increasing sample points.

        Returns the one-dimensional piecewise linear interpolant to a function
        with given data points :math:`(xp, fp)`, evaluated at :math:`x`

        Args:
            x: the :math:`x`-coordinates at which to evaluate the interpolated
                values.
            xp: the :math:`x`-coordinates of data points, must be increasing.
            fp: the :math:`y`-coordinates of data points, same length as `xp`.

        Returns:
            the interpolated values, same size as `x`.
        """
        original_shape = x.shape
        x = x.flatten()
        xp = xp.flatten()
        fp = fp.flatten()

        m = (fp[1:] - fp[:-1]) / (xp[1:] - xp[:-1])  # slope
        b = fp[:-1] - (m * xp[:-1])

        indices = torch.searchsorted(xp, x, right=False) - 1

        interpolated = m[indices] * x + b[indices]

        return interpolated.reshape(original_shape)

    def L2PNR(
        self,
        v: BatchTensor,
        eta: BatchTensor,
    ) -> BatchTensor:
        eta2 = eta**2
        x = v**2
        x2 = x**2
        tmp = (
            eta
            * (
                1.0
                + (1.5 + eta / 6.0) * x
                + (3.375 - (19.0 * eta) / 8.0 - eta2 / 24.0) * x2
            )
        ) / x**0.5

        return tmp

    def convert_spins(
        self,
        m1: BatchTensor,
        m2: BatchTensor,
        f_ref: float,
        phic: BatchTensor,
        inclination: BatchTensor,
        s1x: BatchTensor,
        s1y: BatchTensor,
        s1z: BatchTensor,
        s2x: BatchTensor,
        s2y: BatchTensor,
        s2z: BatchTensor,
    ) -> Tuple[
        BatchTensor,
        BatchTensor,
        BatchTensor,
        BatchTensor,
        BatchTensor,
        BatchTensor,
        BatchTensor,
    ]:
        M = m1 + m2
        m1_2 = m1 * m1
        m2_2 = m2 * m2
        eta = m1 * m2 / (M * M)  # Symmetric mass-ratio

        # From the components in the source frame, we can easily determine
        # chi1_l, chi2_l, chip and phi_aligned, which we need to return.
        # We also compute the spherical angles of J,
        # which we need to transform to the J frame

        # Aligned spins
        chi1_l = s1z  # Dimensionless aligned spin on BH 1
        chi2_l = s2z  # Dimensionless aligned spin on BH 2

        # Magnitude of the spin projections in the orbital plane
        S1_perp = m1_2 * torch.sqrt(s1x**2 + s1y**2)
        S2_perp = m2_2 * torch.sqrt(s2x**2 + s2y**2)

        A1 = 2 + (3 * m2) / (2 * m1)
        A2 = 2 + (3 * m1) / (2 * m2)
        ASp1 = A1 * S1_perp
        ASp2 = A2 * S2_perp
        num = torch.maximum(ASp1, ASp2)
        den = A2 * m2_2  # warning: this assumes m2 > m1
        chip = num / den

        m_sec = M * MTSUN_SI
        piM = PI * m_sec
        v_ref = (piM * f_ref) ** (1 / 3)
        L0 = M * M * self.L2PNR(v_ref, eta)
        J0x_sf = m1_2 * s1x + m2_2 * s2x
        J0y_sf = m1_2 * s1y + m2_2 * s2y
        J0z_sf = L0 + m1_2 * s1z + m2_2 * s2z
        J0 = torch.sqrt(J0x_sf * J0x_sf + J0y_sf * J0y_sf + J0z_sf * J0z_sf)

        thetaJ_sf = torch.arccos(J0z_sf / J0)

        phiJ_sf = torch.arctan2(J0y_sf, J0x_sf)

        phi_aligned = -phiJ_sf

        # First we determine kappa
        # in the source frame, the components of N are given in
        # Eq (35c) of T1500606-v6
        Nx_sf = torch.sin(inclination) * torch.cos(PI / 2.0 - phic)
        Ny_sf = torch.sin(inclination) * torch.sin(PI / 2.0 - phic)
        Nz_sf = torch.cos(inclination)

        tmp_x = Nx_sf
        tmp_y = Ny_sf
        tmp_z = Nz_sf

        tmp_x, tmp_y, tmp_z = rotate_z(-phiJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_y(-thetaJ_sf, tmp_x, tmp_y, tmp_z)

        kappa = -torch.arctan2(tmp_y, tmp_x)

        # Then we determine alpha0, by rotating LN
        tmp_x, tmp_y, tmp_z = 0, 0, 1
        tmp_x, tmp_y, tmp_z = rotate_z(-phiJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_y(-thetaJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_z(kappa, tmp_x, tmp_y, tmp_z)

        alpha0 = torch.arctan2(tmp_y, tmp_x)

        # Finally we determine thetaJ, by rotating N
        tmp_x, tmp_y, tmp_z = Nx_sf, Ny_sf, Nz_sf
        tmp_x, tmp_y, tmp_z = rotate_z(-phiJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_y(-thetaJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_z(kappa, tmp_x, tmp_y, tmp_z)
        Nx_Jf, Nz_Jf = tmp_x, tmp_z
        thetaJN = torch.arccos(Nz_Jf)

        # Finally, we need to redefine the polarizations:
        # PhenomP's polarizations are defined following Arun et al
        # (arXiv:0810.5336)
        # i.e. projecting the metric onto the P,Q,N triad defined with
        # P=NxJ/|NxJ| (see (2.6) in there).
        # By contrast, the triad X,Y,N used in LAL
        # ("waveframe" in the nomenclature of T1500606-v6)
        # is defined in e.g. eq (35) of this document
        # (via its components in the source frame;
        # note we use the default Omega=Pi/2).
        # Both triads differ from each other by a rotation around N by an angle
        # \zeta and we need to rotate the polarizations accordingly by 2\zeta

        Xx_sf = -torch.cos(inclination) * torch.sin(phic)
        Xy_sf = -torch.cos(inclination) * torch.cos(phic)
        Xz_sf = torch.sin(inclination)
        tmp_x, tmp_y, tmp_z = Xx_sf, Xy_sf, Xz_sf
        tmp_x, tmp_y, tmp_z = rotate_z(-phiJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_y(-thetaJ_sf, tmp_x, tmp_y, tmp_z)
        tmp_x, tmp_y, tmp_z = rotate_z(kappa, tmp_x, tmp_y, tmp_z)

        # Now the tmp_a are the components of X in the J frame
        # We need the polar angle of that vector in the P,Q basis of Arun et al
        # P = NxJ/|NxJ| and since we put N in the (pos x)z half plane of the J
        # frame
        PArunx_Jf = 0.0
        PAruny_Jf = -1.0
        PArunz_Jf = 0.0

        # Q = NxP
        QArunx_Jf = Nz_Jf
        QAruny_Jf = 0.0
        QArunz_Jf = -Nx_Jf

        # Calculate the dot products XdotPArun and XdotQArun
        XdotPArun = tmp_x * PArunx_Jf + tmp_y * PAruny_Jf + tmp_z * PArunz_Jf
        XdotQArun = tmp_x * QArunx_Jf + tmp_y * QAruny_Jf + tmp_z * QArunz_Jf

        zeta_polariz = torch.arctan2(XdotQArun, XdotPArun)
        return chi1_l, chi2_l, chip, thetaJN, alpha0, phi_aligned, zeta_polariz

    # TODO: add input and output types
    def SpinWeightedY(self, theta, phi, s, l, m):  # noqa: E741
        "copied from SphericalHarmonics.c in LAL"
        if s == -2:
            if l == 2:  # noqa: E741
                if m == -2:
                    fac = (
                        torch.sqrt(torch.tensor(5.0 / (64.0 * PI)))
                        * (1.0 - torch.cos(theta))
                        * (1.0 - torch.cos(theta))
                    )
                elif m == -1:
                    fac = (
                        torch.sqrt(torch.tensor(5.0 / (16.0 * PI)))
                        * torch.sin(theta)
                        * (1.0 - torch.cos(theta))
                    )
                elif m == 0:
                    fac = (
                        torch.sqrt(torch.tensor(15.0 / (32.0 * PI)))
                        * torch.sin(theta)
                        * torch.sin(theta)
                    )
                elif m == 1:
                    fac = (
                        torch.sqrt(torch.tensor(5.0 / (16.0 * PI)))
                        * torch.sin(theta)
                        * (1.0 + torch.cos(theta))
                    )
                elif m == 2:
                    fac = (
                        torch.sqrt(torch.tensor(5.0 / (64.0 * PI)))
                        * (1.0 + torch.cos(theta))
                        * (1.0 + torch.cos(theta))
                    )
                else:
                    raise ValueError(
                        f"Invalid mode s={s}, l={l}, m={m} - require |m| <= l"
                    )
                return fac * torch.complex(
                    torch.cos(torch.tensor(m * phi)),
                    torch.sin(torch.tensor(m * phi)),
                )

    def WignerdCoefficients(
        self,
        v: BatchTensor,
        SL: BatchTensor,
        eta: BatchTensor,
        Sp: BatchTensor,
    ) -> Tuple[BatchTensor, BatchTensor]:
        # We define the shorthand s := Sp / (L + SL)
        L = self.L2PNR(v, eta)
        s = (Sp / (L + SL)).mT
        s2 = s**2
        cos_beta = 1.0 / (1.0 + s2) ** 0.5
        cos_beta_half = ((1.0 + cos_beta) / 2.0) ** 0.5  # cos(beta/2)
        sin_beta_half = ((1.0 - cos_beta) / 2.0) ** 0.5  # sin(beta/2)

        return cos_beta_half, sin_beta_half

    def ComputeNNLOanglecoeffs(
        self,
        q: BatchTensor,
        chil: BatchTensor,
        chip: BatchTensor,
    ) -> Dict[str, BatchTensor]:
        m2 = q / (1.0 + q)
        m1 = 1.0 / (1.0 + q)
        dm = m1 - m2
        mtot = 1.0
        eta = m1 * m2  # mtot = 1
        eta2 = eta * eta
        eta3 = eta2 * eta
        eta4 = eta3 * eta
        mtot2 = mtot * mtot
        mtot4 = mtot2 * mtot2
        mtot6 = mtot4 * mtot2
        mtot8 = mtot6 * mtot2
        chil2 = chil * chil
        chip2 = chip * chip
        chip4 = chip2 * chip2
        dm2 = dm * dm
        dm3 = dm2 * dm
        m2_2 = m2 * m2
        m2_3 = m2_2 * m2
        m2_4 = m2_3 * m2
        m2_5 = m2_4 * m2
        m2_6 = m2_5 * m2
        m2_7 = m2_6 * m2
        m2_8 = m2_7 * m2

        angcoeffs = {}
        angcoeffs["alphacoeff1"] = -0.18229166666666666 - (5 * dm) / (
            64.0 * m2
        )

        angcoeffs["alphacoeff2"] = (-15 * dm * m2 * chil) / (
            128.0 * mtot2 * eta
        ) - (35 * m2_2 * chil) / (128.0 * mtot2 * eta)

        angcoeffs["alphacoeff3"] = (
            -1.7952473958333333
            - (4555 * dm) / (7168.0 * m2)
            - (15 * chip2 * dm * m2_3) / (128.0 * mtot4 * eta2)
            - (35 * chip2 * m2_4) / (128.0 * mtot4 * eta2)
            - (515 * eta) / 384.0
            - (15 * dm2 * eta) / (256.0 * m2_2)
            - (175 * dm * eta) / (256.0 * m2)
        )

        angcoeffs["alphacoeff4"] = (
            -(35 * PI) / 48.0
            - (5 * dm * PI) / (16.0 * m2)
            + (5 * dm2 * chil) / (16.0 * mtot2)
            + (5 * dm * m2 * chil) / (3.0 * mtot2)
            + (2545 * m2_2 * chil) / (1152.0 * mtot2)
            - (5 * chip2 * dm * m2_5 * chil) / (128.0 * mtot6 * eta3)
            - (35 * chip2 * m2_6 * chil) / (384.0 * mtot6 * eta3)
            + (2035 * dm * m2 * chil) / (21504.0 * mtot2 * eta)
            + (2995 * m2_2 * chil) / (9216.0 * mtot2 * eta)
        )

        angcoeffs["alphacoeff5"] = (
            4.318908476114694
            + (27895885 * dm) / (2.1676032e7 * m2)
            - (15 * chip4 * dm * m2_7) / (512.0 * mtot8 * eta4)
            - (35 * chip4 * m2_8) / (512.0 * mtot8 * eta4)
            - (485 * chip2 * dm * m2_3) / (14336.0 * mtot4 * eta2)
            + (475 * chip2 * m2_4) / (6144.0 * mtot4 * eta2)
            + (15 * chip2 * dm2 * m2_2) / (256.0 * mtot4 * eta)
            + (145 * chip2 * dm * m2_3) / (512.0 * mtot4 * eta)
            + (575 * chip2 * m2_4) / (1536.0 * mtot4 * eta)
            + (39695 * eta) / 86016.0
            + (1615 * dm2 * eta) / (28672.0 * m2_2)
            - (265 * dm * eta) / (14336.0 * m2)
            + (955 * eta2) / 576.0
            + (15 * dm3 * eta2) / (1024.0 * m2_3)
            + (35 * dm2 * eta2) / (256.0 * m2_2)
            + (2725 * dm * eta2) / (3072.0 * m2)
            - (15 * dm * m2 * PI * chil) / (16.0 * mtot2 * eta)
            - (35 * m2_2 * PI * chil) / (16.0 * mtot2 * eta)
            + (15 * chip2 * dm * m2_7 * chil2) / (128.0 * mtot8 * eta4)
            + (35 * chip2 * m2_8 * chil2) / (128.0 * mtot8 * eta4)
            + (375 * dm2 * m2_2 * chil2) / (256.0 * mtot4 * eta)
            + (1815 * dm * m2_3 * chil2) / (256.0 * mtot4 * eta)
            + (1645 * m2_4 * chil2) / (192.0 * mtot4 * eta)
        )

        angcoeffs["epsiloncoeff1"] = -0.18229166666666666 - (5 * dm) / (
            64.0 * m2
        )
        angcoeffs["epsiloncoeff2"] = (-15 * dm * m2 * chil) / (
            128.0 * mtot2 * eta
        ) - (35 * m2_2 * chil) / (128.0 * mtot2 * eta)
        angcoeffs["epsiloncoeff3"] = (
            -1.7952473958333333
            - (4555 * dm) / (7168.0 * m2)
            - (515 * eta) / 384.0
            - (15 * dm2 * eta) / (256.0 * m2_2)
            - (175 * dm * eta) / (256.0 * m2)
        )
        angcoeffs["epsiloncoeff4"] = (
            -(35 * PI) / 48.0
            - (5 * dm * PI) / (16.0 * m2)
            + (5 * dm2 * chil) / (16.0 * mtot2)
            + (5 * dm * m2 * chil) / (3.0 * mtot2)
            + (2545 * m2_2 * chil) / (1152.0 * mtot2)
            + (2035 * dm * m2 * chil) / (21504.0 * mtot2 * eta)
            + (2995 * m2_2 * chil) / (9216.0 * mtot2 * eta)
        )
        angcoeffs["epsiloncoeff5"] = (
            4.318908476114694
            + (27895885 * dm) / (2.1676032e7 * m2)
            + (39695 * eta) / 86016.0
            + (1615 * dm2 * eta) / (28672.0 * m2_2)
            - (265 * dm * eta) / (14336.0 * m2)
            + (955 * eta2) / 576.0
            + (15 * dm3 * eta2) / (1024.0 * m2_3)
            + (35 * dm2 * eta2) / (256.0 * m2_2)
            + (2725 * dm * eta2) / (3072.0 * m2)
            - (15 * dm * m2 * PI * chil) / (16.0 * mtot2 * eta)
            - (35 * m2_2 * PI * chil) / (16.0 * mtot2 * eta)
            + (375 * dm2 * m2_2 * chil2) / (256.0 * mtot4 * eta)
            + (1815 * dm * m2_3 * chil2) / (256.0 * mtot4 * eta)
            + (1645 * m2_4 * chil2) / (192.0 * mtot4 * eta)
        )
        return angcoeffs

    def FinalSpin_inplane(
        self,
        m1: BatchTensor,
        m2: BatchTensor,
        chi1_l: BatchTensor,
        chi2_l: BatchTensor,
        chip: BatchTensor,
    ) -> BatchTensor:
        M = m1 + m2
        eta = m1 * m2 / (M * M)
        eta2 = eta * eta
        # m1 > m2, the convention used in phenomD
        # (not the convention of internal phenomP)
        q_factor = m1 / M
        af_parallel = self.FinalSpin0815(eta, eta2, chi1_l, chi2_l)
        Sperp = chip * q_factor * q_factor
        af = torch.copysign(
            torch.ones_like(af_parallel), af_parallel
        ) * torch.sqrt(Sperp * Sperp + af_parallel * af_parallel)
        return af

    def phP_get_fRD_fdamp(
        self, m1, m2, chi1_l, chi2_l, chip
    ) -> Tuple[BatchTensor, BatchTensor]:
        # m1 > m2 should hold here
        finspin = self.FinalSpin_inplane(m1, m2, chi1_l, chi2_l, chip)
        m1_s = m1 * MTSUN_SI
        m2_s = m2 * MTSUN_SI
        M_s = m1_s + m2_s
        eta_s = m1_s * m2_s / (M_s**2.0)
        eta_s2 = eta_s * eta_s
        Erad = self.PhenomInternal_EradRational0815(
            eta_s, eta_s2, chi1_l, chi2_l
        )
        fRD = self.interpolate(finspin, self.qnmdata_a, self.qnmdata_fring) / (
            1.0 - Erad
        )
        fdamp = self.interpolate(
            finspin, self.qnmdata_a, self.qnmdata_fdamp
        ) / (1.0 - Erad)
        return fRD / M_s, fdamp / M_s

    def get_Amp0(self, fM_s: BatchTensor, eta: BatchTensor) -> BatchTensor:
        Amp0 = (
            (2.0 / 3.0 * eta.unsqueeze(1)) ** (1.0 / 2.0)
            * (fM_s) ** (-7.0 / 6.0)
            * PI ** (-1.0 / 6.0)
        )
        return Amp0



import torch
from jaxtyping import Float
from torch import Tensor


class SignalInverter(torch.nn.Module):
    """
    Takes a tensor of timeseries of arbitrary dimension
    and randomly inverts (i.e. h(t) -> -h(t))
    each timeseries with probability `prob`.

    Args:
        prob:
            Probability that a timeseries is inverted
    """

    def __init__(self, prob: float = 0.5):
        super().__init__()
        self.prob = prob

    def forward(
        self, X: Float[Tensor, "*batch time"]
    ) -> Float[Tensor, "*batch time"]:
        mask = torch.rand(size=X.shape[:-1]) < self.prob
        X[mask] *= -1
        return X


class SignalReverser(torch.nn.Module):
    """
    Takes a tensor of timeseries of arbitrary dimension
    and randomly reverses (i.e. h(t) -> h(-t))
    each timeseries with probability `prob`.

    Args:
        prob:
            Probability that a kernel is reversed
    """

    def __init__(self, prob: float = 0.5):
        super().__init__()
        self.prob = prob

    def forward(
        self, X: Float[Tensor, "*batch time"]
    ) -> Float[Tensor, "*batch time"]:
        mask = torch.rand(size=X.shape[:-1]) < self.prob
        X[mask] = X[mask].flip(-1)
        return X



"""
Tools for manipulating raw gravitational waveforms
and projecting them onto interferometer responses.
Much of the projection code is an extension of the
implementation made available in bilby:

https://arxiv.org/abs/1811.02042

Specifically the code here:
https://github.com/lscsoft/bilby/blob/master/bilby/gw/detector/interferometer.py
"""

from typing import List, Tuple, Union

import torch
from jaxtyping import Float
from torch import Tensor

from .constants import C
from .types import (
    BatchTensor,
    NetworkDetectorTensors,
    NetworkVertices,
    PSDTensor,
    TensorGeometry,
    VectorGeometry,
    WaveformTensor,
)
from .utils.interferometer import InterferometerGeometry


def outer(x: VectorGeometry, y: VectorGeometry) -> TensorGeometry:
    """
    Compute the outer product of two batches of vectors
    """
    return torch.einsum("...i,...j->...ij", x, y)


def plus(m: VectorGeometry, n: VectorGeometry) -> TensorGeometry:
    return outer(m, m) - outer(n, n)


def cross(m: VectorGeometry, n: VectorGeometry) -> TensorGeometry:
    return outer(m, n) + outer(n, m)


def breathing(m: VectorGeometry, n: VectorGeometry) -> TensorGeometry:
    return outer(m, m) + outer(n, n)


polarization_funcs = {
    "plus": plus,
    "cross": cross,
    "breathing": breathing,
}


def compute_antenna_responses(
    theta: BatchTensor,
    psi: BatchTensor,
    phi: BatchTensor,
    detector_tensors: NetworkDetectorTensors,
    modes: List[str],
) -> Float[Tensor, "batch polarizations num_ifos"]:
    """
    Compute the antenna pattern factors of a batch of
    waveforms as a function of the sky parameters of
    their sources as well as the detector tensors of
    the interferometers whose response is being
    calculated.

    Args:
        theta:
            Angle of each source in radians relative
            to the celestial equator
        psi:
            Angle in radians between each source's
            natural polarization basis and the basis
            which has the 0th unit vector pointing along
            the celestial equator
        phi:
            Angle in radians between each source's right
            ascension and the right ascension of the
            geocenter
        detector_tensors:
            Detector tensor for each of the interferometers
            for which a response is being calculated, stacked
            along the 0th axis
        modes:
            Which polarization modes to compute the response for
    Returns:
        A tensor representing interferometer antenna pattern
            factors for each of the polarizations of each of
            the waveforms, for each interferometer.
    """

    # add a dimension so that we can do some tensor
    # manipulations batch-wise later
    theta = theta.view(-1, 1)
    psi = psi.view(-1, 1)
    phi = phi.view(-1, 1)

    # pre-compute all our trigonometric functions
    # since we'll end up using them all at least twice
    sin_theta, cos_theta = torch.sin(theta), torch.cos(theta)
    sin_psi, cos_psi = torch.sin(psi), torch.cos(psi)
    sin_phi, cos_phi = torch.sin(phi), torch.cos(phi)

    # u and v are the unit vectors orthogonal to the
    # source's propogation vector, where v is
    # parallel to the celestial equator
    u = torch.cat(
        [cos_phi * cos_theta, cos_theta * sin_phi, -sin_theta], axis=1
    )
    v = torch.cat([-sin_phi, cos_phi, torch.zeros_like(phi)], axis=1)

    # m and n are the unit vectors along the waveforms
    # natural polarization basis
    m = -u * sin_psi - v * cos_psi
    n = -u * cos_psi + v * sin_psi

    # compute the polarization tensor of each signal
    # as a function of the parameters of its source
    polarizations = []
    for mode in modes:
        try:
            polarization = polarization_funcs[mode](m, n)
        except KeyError:
            raise ValueError(f"No polarization mode {mode}")

        # add a dummy dimension for concatenating
        polarizations.append(polarization)

    # shape: batch x num_polarizations x 3 x 3
    polarization = torch.stack(polarizations, axis=1)

    # compute the weight of each interferometer's response
    # to each polarization: batch x polarizations x ifos
    return torch.einsum("...jk,ijk->...i", polarization, detector_tensors)


def shift_responses(
    responses: WaveformTensor,
    theta: BatchTensor,
    phi: BatchTensor,
    vertices: NetworkVertices,
    sample_rate: float,
) -> WaveformTensor:
    omega = torch.column_stack(
        [
            torch.sin(theta) * torch.cos(phi),
            torch.sin(theta) * torch.sin(phi),
            torch.cos(theta),
        ]
    ).view(-1, 1, 3)

    # compute the time delay between the geocenter
    # and each interferometer in units of s, then
    # convert to units of samples and discretize.
    # Divide by c in the second line so that we only
    # need to multiply the array by a single float
    dt = -(omega * vertices).sum(axis=-1)
    dt *= sample_rate / C
    dt = torch.trunc(dt).type(torch.int64)

    # rolling by gathering implementation based on
    # https://stackoverflow.com/a/68641864
    # start by just creating a big arange along the last axis
    idx = torch.ones_like(responses).type(torch.int64)
    idx = torch.cumsum(idx, axis=-1) - 1

    # apply the offset to the indices along the last axis,
    # then modulo by the waveform size to make all the
    # specified indices legitimate and unique
    idx -= dt[:, :, None]
    idx %= idx.shape[-1]

    # unfortunately I can't figure out how to do this
    # last step without doing looping over the ifos
    rolled = []
    for i in range(len(vertices)):
        ifo = torch.gather(responses[:, i], 1, idx[:, i])
        rolled.append(ifo)
    return torch.stack(rolled, axis=1)


def compute_observed_strain(
    dec: BatchTensor,
    psi: BatchTensor,
    phi: BatchTensor,
    detector_tensors: NetworkDetectorTensors,
    detector_vertices: NetworkVertices,
    sample_rate: float,
    **polarizations: Float[Tensor, "batch time"],
) -> WaveformTensor:
    """
    Compute the strain timeseries $h(t)$ observed by a network
    of interferometers from the given polarization timeseries
    corresponding to gravitational waveforms from sources with
    the indicated sky parameters.

    Args:
        dec:
            Declination of each source in radians relative
            to the celestial north
        psi:
            Angle in radians between each source's
            natural polarization basis and the basis
            which has the 0th unit vector pointing along
            the celestial equator
        phi:
            Angle in radians between each source's right
            ascension and the right ascension of the
            geocenter
        detector_tensors:
            Detector tensor for each of the interferometers
            for which observed strain is being calculated,
            stacked along the 0th axis
        detector_vertices:
            Vertices for each interferometer's spatial location
            relative to the geocenter. Used to compute delay
            between the waveform observed at the geocenter and
            the one observed at the detector site. To avoid
            adding any delay between the two, reset your coordinates
            such that the desired interferometer is at `(0., 0., 0.)`.
        sample_rate:
            Rate at which the polarization timeseries have been sampled
        polarziations:
            Timeseries for each waveform polarization which
            contributes to the interferometer response. Allowed
            polarizations are `cross`, `plus`, and `breathing`.
    Returns:
        Tensor representing the observed strain at each
        interferometer for each waveform.
    """

    # TODO: just use theta as the input parameter?
    # note that ** syntax is ordered, so we're safe
    # to be lazy and use `list` for the keys and values
    theta = torch.pi / 2 - dec
    antenna_responses = compute_antenna_responses(
        theta, psi, phi, detector_tensors, list(polarizations)
    )

    polarizations = torch.stack(list(polarizations.values()), axis=1)
    waveforms = torch.einsum(
        "...pi,...pt->...it", antenna_responses, polarizations
    )

    return shift_responses(
        waveforms, theta, phi, detector_vertices, sample_rate
    )


def get_ifo_geometry(
    *ifos: str,
) -> Tuple[NetworkDetectorTensors, NetworkVertices]:
    """
    For a given list of interferometer names, retrieve and
    concatenate the associated detector tensors and vertices
    of those interferometers.

    Args:
        ifos: Names of the interferometers whose geometry to retrieve
    Returns:
        A concatenation of the detector tensors of each interferometer
        A concatenation of the vertices of each interferometer
    """

    tensors, vertices = [], []
    for ifo in ifos:
        ifo = InterferometerGeometry(ifo)
        detector_tensor = plus(ifo.x_arm, ifo.y_arm) / 2
        tensors.append(detector_tensor)
        vertices.append(ifo.vertex)

    tensors = torch.stack(tensors)
    vertices = torch.stack(vertices)
    return torch.Tensor(tensors), torch.Tensor(vertices)


def compute_ifo_snr(
    responses: WaveformTensor,
    psd: PSDTensor,
    sample_rate: float,
    highpass: Union[float, Float[Tensor, " frequency"], None] = None,
    lowpass: Union[float, Float[Tensor, " frequency"], None] = None,
) -> Float[Tensor, "batch num_ifos"]:
    r"""Compute the SNRs of a batch of interferometer responses

    Compute the signal to noise ratio (SNR) of individual
    interferometer responses to gravitational waveforms with
    respect to a background PSD for each interferometer. The
    SNR of the $i$th waveform at the $j$th interferometer
    is computed as:

    $$\rho_{ij} =
        4 \int_{f_{\text{min}}}^{f_{\text{max}}}
        \frac{\tilde{h_{ij}}(f)\tilde{h_{ij}}^*(f)}
        {S_n^{(j)}(f)}df$$

    Where $f_{\text{min}}$ is a minimum frequency denoted
    by `highpass`, `f_{\text{max}}` is the maximum frequency
    denoted by `lowpass`, which defaults to the Nyquist frequency
    dictated by `sample_rate`; `\tilde{h_{ij}}` and `\tilde{h_{ij}}*`
    indicate the fourier transform of the $i$th waveform at
    the $j$th inteferometer and its complex conjugate, respectively;
    and $S_n^{(j)}$ is the backround PSD at the $j$th interferometer.

    Args:
        responses:
            A batch of interferometer responses to a batch of
            raw gravitational waveforms
        psd:
            The one-sided power spectral density of the background
            noise at each interferometer to which a response
            in `responses` has been calculated. If 2D, each row of
            `psd` will be assumed to be the background PSD for each
            channel of _every_ batch element in `responses`. If 3D,
            this should contain a background PSD for each channel
            of each element in `responses`, and therefore the first
            two dimensions of `psd` and `responses` should match.
        sample_rate:
            The frequency at which the waveform responses timeseries
            have been sampled. Upon fourier transforming, should
            match the frequency resolution of the provided PSDs.
        highpass:
            The minimum frequency above which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out low frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies up to `lowpass`
            will contribute to the SNR calculation.
        lowpass:
            The maximum frequency below which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out high frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies from `highpass` up to
            the Nyquist freqyency will contribute to the SNR calculation.
    Returns:
        Batch of SNRs computed for each interferometer
    """

    # TODO: should we do windowing here?
    # compute frequency power, upsampling precision so that
    # computing absolute value doesn't accidentally zero some
    # values out.
    fft = torch.fft.rfft(responses, axis=-1).type(torch.complex128)
    fft = fft.abs() / sample_rate

    # divide by background asd, then go back to FP32 precision
    # and square now that values are back in a reasonable range
    integrand = fft / (psd**0.5)
    integrand = integrand.type(torch.float32) ** 2

    # mask out frequency components if a critical
    # frequency or frequency mask was provided
    if highpass is not None:
        if not isinstance(highpass, torch.Tensor):
            freqs = torch.fft.rfftfreq(responses.shape[-1], 1 / sample_rate)
            highpass = freqs >= highpass
        elif len(highpass) != integrand.shape[-1]:
            raise ValueError(
                "Can't apply highpass filter mask with {} frequency bins"
                "to signal fft with {} frequency bins".format(
                    len(highpass), integrand.shape[-1]
                )
            )
        integrand *= highpass.to(integrand.device)
    if lowpass is not None:
        if not isinstance(lowpass, torch.Tensor):
            freqs = torch.fft.rfftfreq(responses.shape[-1], 1 / sample_rate)
            lowpass = freqs < lowpass
        elif len(lowpass) != integrand.shape[-1]:
            raise ValueError(
                "Can't apply lowpass filter mask with {} frequency bins"
                "to signal fft with {} frequency bins".format(
                    len(lowpass), integrand.shape[-1]
                )
            )
        integrand *= lowpass.to(integrand.device)

    # sum over the desired frequency range and multiply
    # by df to turn it into an integration (and get
    # our units to drop out)
    # TODO: we could in principle do this without requiring
    # that the user specify the sample rate by taking the
    # fft as-is (without dividing by sample rate) and then
    # taking the mean here (or taking the sum and dividing
    # by the sum of `highpass` if it's a mask). If we want
    # to allow the user to pass a float for highpass, we'll
    # need the sample rate to compute the mask, but if we
    # replace this with a `mask` argument instead we're in
    # the clear
    df = sample_rate / responses.shape[-1]
    integrated = integrand.sum(axis=-1) * df

    # multiply by 4 for mystical reasons
    integrated = 4 * integrated  # rho-squared
    return torch.sqrt(integrated)


def compute_network_snr(
    responses: WaveformTensor,
    psd: PSDTensor,
    sample_rate: float,
    highpass: Union[float, Float[Tensor, " frequency"], None] = None,
    lowpass: Union[float, Float[Tensor, " frequency"], None] = None,
) -> BatchTensor:
    r"""
    Compute the total SNR from a gravitational waveform
    from a network of interferometers. The total SNR for
    the $i$th waveform is computed as

    $$\rho_i = \sqrt{\sum_{j}^{N}\rho_{ij}^2}$$

    where \rho_{ij} is the SNR for the $i$th waveform at
    the $j$th interferometer in the network and $N$ is
    the total number of interferometers.

    Args:
        responses:
            A batch of interferometer responses to a batch of
            raw gravitational waveforms
        backgrounds:
            The one-sided power spectral density of the background
            noise at each interferometer to which a response
            in `responses` has been calculated. If 2D, each row of
            `psd` will be assumed to be the background PSD for each
            channel of _every_ batch element in `responses`. If 3D,
            this should contain a background PSD for each channel
            of each element in `responses`, and therefore the first
            two dimensions of `psd` and `responses` should match.
        sample_rate:
            The frequency at which the waveform responses timeseries
            have been sampled. Upon fourier transforming, should
            match the frequency resolution of the provided PSDs.
        highpass:
            The minimum frequency above which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out low frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies up to `sample_rate / 2`
            will contribute to the SNR calculation.
        lowpass:
            The maximum frequency below which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out high frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies from `highpass` up to
            the Nyquist freqyency will contribute to the SNR calculation.
    Returns:
        Batch of SNRs for each waveform across the interferometer network
    """
    snrs = compute_ifo_snr(responses, psd, sample_rate, highpass, lowpass)
    snrs = snrs**2
    return snrs.sum(axis=-1) ** 0.5


def reweight_snrs(
    responses: WaveformTensor,
    target_snrs: Union[float, BatchTensor],
    psd: PSDTensor,
    sample_rate: float,
    highpass: Union[float, Float[Tensor, " frequency"], None] = None,
    lowpass: Union[float, Float[Tensor, " frequency"], None] = None,
) -> WaveformTensor:
    """Scale interferometer responses such that they have a desired SNR

    Args:
        responses:
            A batch of interferometer responses to a batch of
            raw gravitational waveforms
        target_snrs:
            Either a tensor of desired SNRs for each waveform,
            or a single SNR to which all waveforms should be scaled.
        psd:
            The one-sided power spectral density of the background
            noise at each interferometer to which a response
            in `responses` has been calculated. If 2D, each row of
            `psd` will be assumed to be the background PSD for each
            channel of _every_ batch element in `responses`. If 3D,
            this should contain a background PSD for each channel
            of each element in `responses`, and therefore the first
            two dimensions of `psd` and `responses` should match.
        sample_rate:
            The frequency at which the waveform responses timeseries
            have been sampled. Upon fourier transforming, should
            match the frequency resolution of the provided PSDs.
        highpass:
            The minimum frequency above which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out low frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies up to `sample_rate / 2`
            will contribute to the SNR calculation.
        lowpass:
            The maximum frequency below which to compute the SNR.
            If a tensor is provided, it will be assumed to be a
            pre-computed mask used to 0-out high frequency components.
            If a float, it will be used to compute such a mask. If
            left as `None`, all frequencies from `highpass` up to
            the Nyquist freqyency will contribute to the SNR calculation.
    Returns:
        Rescaled interferometer responses
    """

    snrs = compute_network_snr(responses, psd, sample_rate, highpass, lowpass)
    weights = target_snrs / snrs
    return responses * weights[:, None, None]



import torch

from ml4gw.constants import C, G
from ml4gw.types import BatchTensor


def taylor_t2_timing_0pn_coeff(total_mass: BatchTensor, eta: BatchTensor):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L1528
    """

    output = total_mass * G / C**3
    return -5.0 * output / (256.0 * eta)


def taylor_t2_timing_2pn_coeff(eta: BatchTensor):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L1545
    """
    return 7.43 / 2.52 + 11.0 / 3.0 * eta


def taylor_t2_timing_4pn_coeff(eta: BatchTensor):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L1560
    """
    return 30.58673 / 5.08032 + 54.29 / 5.04 * eta + 61.7 / 7.2 * eta**2


def taylor_t3_frequency_0pn_coeff(total_mass: BatchTensor):
    """
    https://git.ligo.org/lscsoft/lalsuite/-/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c#L1723
    """
    output = total_mass * G / C**3.0
    return 1.0 / (8.0 * torch.pi * output)



from typing import Optional

import torch

from ..gw import compute_network_snr
from ..types import BatchTensor, TimeSeries2d, WaveformTensor
from .transform import FittableSpectralTransform


class SnrRescaler(FittableSpectralTransform):
    def __init__(
        self,
        num_channels: int,
        sample_rate: float,
        waveform_duration: float,
        highpass: Optional[float] = None,
        lowpass: Optional[float] = None,
        dtype: torch.dtype = torch.float32,
    ) -> None:
        super().__init__()
        self.sample_rate = sample_rate
        self.num_channels = num_channels

        waveform_size = int(waveform_duration * sample_rate)
        num_freqs = int(waveform_size // 2 + 1)

        buff = torch.zeros((num_channels, num_freqs), dtype=dtype)
        self.register_buffer("background", buff)

        if highpass is not None:
            freqs = torch.fft.rfftfreq(waveform_size, 1 / sample_rate)
            self.register_buffer(
                "highpass_mask", freqs >= highpass, persistent=False
            )
        else:
            self.highpass_mask = None
        if lowpass is not None:
            freqs = torch.fft.rfftfreq(waveform_size, 1 / sample_rate)
            self.register_buffer(
                "lowpass_mask", freqs < lowpass, persistent=False
            )
        else:
            self.lowpass_mask = None

    def fit(
        self,
        *background: TimeSeries2d,
        fftlength: Optional[float] = None,
        overlap: Optional[float] = None,
    ):
        if len(background) != self.num_channels:
            raise ValueError(
                "Expected to fit whitening transform on {} background "
                "timeseries, but was passed {}".format(
                    self.num_channels, len(background)
                )
            )

        num_freqs = self.background.size(1)
        psds = []
        for x in background:
            psd = self.normalize_psd(
                x, self.sample_rate, num_freqs, fftlength, overlap
            )
            psds.append(psd)
        background = torch.stack(psds)
        super().build(background=background)

    def forward(
        self,
        responses: WaveformTensor,
        target_snrs: Optional[BatchTensor] = None,
    ):
        snrs = compute_network_snr(
            responses,
            self.background,
            self.sample_rate,
            self.highpass_mask,
            self.lowpass_mask,
        )
        if target_snrs is None:
            idx = torch.randperm(len(snrs))
            target_snrs = snrs[idx]

        weights = target_snrs / snrs
        rescaled_responses = responses * weights.view(-1, 1, 1)

        return rescaled_responses, target_snrs



"""
Adaptation of code from https://github.com/dottormale/Qtransform
"""

from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor


class SplineInterpolate(torch.nn.Module):
    """
    Perform 1D or 2D spline interpolation based on De Boor's method.
    Supports batched, multi-channel inputs, so acceptable data
    shapes are `(width)`, `(height, width)`, `(batch, width)`,
    `(batch, height, width)`, `(batch, channel, width)`, and
    `(batch, channel, height, width)`.

    During initialization of this Module, both the desired input
    and output coordinate Tensors can be specified to allow
    pre-computation of the B-spline basis matrices, though the only
    mandatory argument is the coordinates of the data along the
    `width` dimension. If no argument is given for coordinates along
    the `height` dimension, it is assumed that 1D interpolation is
    desired.

    Unlike scipy's implementation of spline interpolation, the data
    to be interpolated is not passed until actually calling the
    object. This is useful for cases where the input and output
    coordinates are known in advance, but the data is not, so that
    the interpolator can be set up ahead of time.

    WARNING: compared to scipy's spline interpolation, this method
    produces edge artifacts when the output coordinates are near
    the boundaries of the input coordinates. Therefore, it is
    recommended to interpolate only to coordinates that are well
    within the input coordinate range. Unfortunately, the specific
    definition of "well within" changes based on the size of the
    data, so some testing may be required to get good results.

    Args:
        x_in:
            Coordinates of the width dimension of the data
        y_in:
            Coordinates of the height dimension of the data. If not
            specified, it is assumed the 1D interpolation is desired,
            and so the default value is a Tensor of length 1
        kx:
            Degree of spline interpolation along the width dimension.
            Default is cubic.
        ky:
            Degree of spline interpolation along the height dimension.
            Default is cubic.
        sx:
            Regularization factor to avoid singularities during matrix
            inversion for interpolation along the width dimension. Not
            to be confused with the `s` parameter in scipy's spline
            methods, which controls the number of knots.
        sy:
            Regularization factor to avoid singularities during matrix
            inversion for interpolation along the height dimension.
        x_out:
            Coordinates for the data to be interpolated to along the
            width dimension. If not specified during initialization,
            this must be specified during the object call.
        y_out:
            Coordinates for the data to be interpolated to along the
            height dimension. If not specified during initialization,
            this must be specified during the object call.

    """

    def __init__(
        self,
        x_in: Tensor,
        y_in: Tensor = Tensor([1]),
        kx: int = 3,
        ky: int = 3,
        sx: float = 0.001,
        sy: float = 0.001,
        x_out: Optional[Tensor] = None,
        y_out: Optional[Tensor] = None,
    ):
        super().__init__()
        self.kx = kx
        self.ky = ky
        self.sx = sx
        self.sy = sy
        self.register_buffer("x_in", x_in)
        self.register_buffer("y_in", y_in)
        self.register_buffer("x_out", x_out)
        self.register_buffer("y_out", y_out)

        tx, Bx, BxT_Bx = self._compute_knots_and_basis_matrices(x_in, kx, sx)
        self.register_buffer("tx", tx)
        self.register_buffer("Bx", Bx)
        self.register_buffer("BxT_Bx", BxT_Bx)

        ty, By, ByT_By = self._compute_knots_and_basis_matrices(y_in, ky, sy)
        self.register_buffer("ty", ty)
        self.register_buffer("By", By)
        self.register_buffer("ByT_By", ByT_By)

        if self.x_out is not None:
            Bx_out = self.bspline_basis_natural(x_out, kx, self.tx)
            self.register_buffer("Bx_out", Bx_out)
        if self.y_out is not None:
            By_out = self.bspline_basis_natural(y_out, ky, self.ty)
            self.register_buffer("By_out", By_out)

    def _compute_knots_and_basis_matrices(self, x, k, s):
        knots = self.generate_natural_knots(x, k)
        basis_matrix = self.bspline_basis_natural(x, k, knots)
        identity = torch.eye(basis_matrix.shape[-1])
        B_T_B = basis_matrix.T @ basis_matrix + s * identity
        return knots, basis_matrix, B_T_B

    def generate_natural_knots(self, x: Tensor, k: int) -> Tensor:
        """
        Generates a natural knot sequence for B-spline interpolation.
        Natural knot sequence means that 2*k knots are added to the beginning
        and end of datapoints as replicas of first and last datapoint
        respectively in order to enforce natural boundary conditions,
        i.e. second derivative = 0.
        The other n nodes are placed in correspondece of the data points.

        Args:
            x: Tensor of data point positions.
            k: Degree of the spline.

        Returns:
            Tensor of knot positions.
        """
        return F.pad(x[None], (k, k), mode="replicate")[0]

    def compute_L_R(
        self,
        x: Tensor,
        t: Tensor,
        d: int,
        m: int,
    ) -> Tuple[Tensor, Tensor]:

        """
        Compute the L and R values for B-spline basis functions.
        L and R are respectively the first and second coefficient multiplying
        B_{i,p-1}(x) and B_{i+1,p-1}(x) in De Boor's recursive formula for
        Bspline basis funciton computation
        See https://en.wikipedia.org/wiki/De_Boor%27s_algorithm for details

        Args:
            x:
                Tensor of data point positions.
            t:
                Tensor of knot positions.
            d:
                Current degree of the basis function.
            m:
                Number of intervals (n - k - 1, where n is the number of knots
                and k is the degree).

        Returns:
            L: Tensor containing left values for the B-spline basis functions.
            R: Tensor containing right values for the B-spline basis functions.
        """
        left_num = x.unsqueeze(1) - t[:m].unsqueeze(0)
        left_den = t[d : m + d] - t[:m]
        L = left_num / left_den.unsqueeze(0)
        L = torch.nan_to_num_(L, nan=0.0, posinf=0.0, neginf=0.0)

        right_num = t[d + 1 : m + d + 1] - x.unsqueeze(1)
        right_den = t[d + 1 : m + d + 1] - t[1 : m + 1]
        R = right_num / right_den.unsqueeze(0)
        R = torch.nan_to_num_(R, nan=0.0, posinf=0.0, neginf=0.0)

        return L, R

    def zeroth_order(
        self,
        x: Tensor,
        k: int,
        t: Tensor,
        n: int,
        m: int,
    ) -> Tensor:

        """
        Compute the zeroth-order B-spline basis functions
        according to de Boors recursive formula.
        See https://en.wikipedia.org/wiki/De_Boor%27s_algorithm for reference

        Args:
            x:
                Tensor of data point positions.
            k:
                Degree of the spline.
            t:
                Tensor of knot positions.
            n:
                Number of data points.
            m:
                Number of intervals (n - k - 1, where n is the number of knots
                and k is the degree).

        Returns:
            b: Tensor containing the zeroth-order B-spline basis functions.
        """
        b = torch.zeros((n, m, k + 1))

        mask_lower = t[: m + 1].unsqueeze(0)[:, :-1] <= x.unsqueeze(1)
        mask_upper = x.unsqueeze(1) < t[: m + 1].unsqueeze(0)[:, 1:]

        b[:, :, 0] = mask_lower & mask_upper
        b[:, 0, 0] = torch.where(x < t[1], torch.ones_like(x), b[:, 0, 0])
        b[:, -1, 0] = torch.where(x >= t[-2], torch.ones_like(x), b[:, -1, 0])
        return b

    def bspline_basis_natural(
        self,
        x: Tensor,
        k: int,
        t: Tensor,
    ) -> Tensor:
        """
        Compute bspline basis function using de Boor's recursive formula
        (See https://en.wikipedia.org/wiki/De_Boor%27s_algorithm for reference)
        Args:
            x: Tensor of data point positions.
            k: Degree of the spline.
            t: Tensor of knot positions.

        Returns:
            Tensor containing the kth-order B-spline basis functions
        """

        if len(x) == 1:
            return torch.eye(1)
        n = x.shape[0]
        m = t.shape[0] - k - 1

        # calculate zeroth order basis funciton
        b = self.zeroth_order(x, k, t, n, m)

        zeros_tensor = torch.zeros(b.shape[0], 1)
        # recursive de Boors formula for bspline basis functions
        for d in range(1, k + 1):
            L, R = self.compute_L_R(x, t, d, m)
            left = L * b[:, :, d - 1]

            temp_b = torch.cat([b[:, 1:, d - 1], zeros_tensor], dim=1)

            right = R * temp_b
            b[:, :, d] = left + right

        return b[:, :, -1]

    def bivariate_spline_fit_natural(self, Z):

        if len(Z.shape) == 3:
            Z_Bx = torch.matmul(Z, self.Bx)
            # ((BxT @ Bx)^-1 @ (Z @ Bx)T)T = Z @ BxT^-1
            return torch.linalg.solve(self.BxT_Bx, Z_Bx.mT).mT

        # Adding batch/channel dimension handling
        # ByT @ Z @ BxW
        ByT_Z_Bx = torch.einsum("ij,bcik,kl->bcjl", self.By, Z, self.Bx)
        # (ByT @ By)^-1 @ (ByT @ Z @ Bx) = By^-1 @ Z @ Bx
        E = torch.linalg.solve(self.ByT_By, ByT_Z_Bx)
        # ((BxT @ Bx)^-1 @ (By^-1 @ Z @ Bx)T)T = By^-1 @ Z @ BxT^-1
        return torch.linalg.solve(self.BxT_Bx, E.mT).mT

    def evaluate_bivariate_spline(self, C: Tensor):
        """
        Evaluate a bivariate spline on a grid of x and y points.

        Args:
            C: Coefficient tensor of shape (batch_size, mx, my).

        Returns:
            Z_interp: Interpolated values at the grid points.
        """
        # Perform matrix multiplication using einsum to get Z_interp
        if len(C.shape) == 3:
            return torch.matmul(C, self.Bx_out.mT)
        return torch.einsum("ik,bckm,mj->bcij", self.By_out, C, self.Bx_out.mT)

    def _validate_inputs(self, Z, x_out, y_out):
        if x_out is None and self.x_out is None:
            raise ValueError(
                "Output x-coordinates were not specified in either object "
                "creation or in forward call"
            )

        if y_out is None and self.y_out is None:
            y_out = self.y_in

        dims = len(Z.shape)
        if dims > 4:
            raise ValueError("Input data has more than 4 dimensions")

        if len(self.y_in) > 1 and dims == 1:
            raise ValueError(
                "An input y-coordinate array with length greater than 1 "
                "was given, but the input data is 1-dimensional. Expected "
                "input data to be at least 2-dimensional"
            )

        # Expand Z to have 4 dimensions
        # There are 6 valid input shapes: (w), (b, w), (b, c, w),
        # (h, w), (b, h, w), and (b, c, h, w).

        # If the input y coordinate array has length 1,
        # assume the first dimension(s) are batch dimensions
        # and that no height dimension is included in Z
        idx = -2 if len(self.y_in) == 1 else -3
        while len(Z.shape) < 4:
            Z = Z.unsqueeze(idx)

        if Z.shape[-2:] != torch.Size([len(self.y_in), len(self.x_in)]):
            raise ValueError(
                "The spatial dimensions of the data tensor do not match "
                "the given input dimensions. "
                f"Expected [{len(self.y_in)}, {len(self.x_in)}], but got "
                f"[{Z.shape[-2]}, {Z.shape[-1]}]"
            )

        return Z, y_out

    def forward(
        self,
        Z: Tensor,
        x_out: Optional[Tensor] = None,
        y_out: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Compute the interpolated data

        Args:
            Z:
                Tensor of data to be interpolated. Must be between 1 and 4
                dimensions. The shape of the tensor must agree with the
                input coordinates given on initialization. If `y_in` was
                not specified during initialization, it is assumed that
                Z does not have a height dimension.
            x_out:
                Coordinates to interpolate the data to along the width
                dimension. Overrides any value that was set during
                initialization.
            y_out:
                Coordinates to interpolate the data to along the height
                dimension. Overrides any value that was set during
                initialization.

        Returns:
            A 4D tensor with shape `(batch, channel, height, width)`.
            Depending on the input data shape, many of these dimensions
            may have length 1.
        """

        Z, y_out = self._validate_inputs(Z, x_out, y_out)

        if x_out is not None:
            self.Bx_out = self.bspline_basis_natural(x_out, self.kx, self.tx)
        if y_out is not None:
            self.By_out = self.bspline_basis_natural(y_out, self.ky, self.ty)

        coef = self.bivariate_spline_fit_natural(Z)
        Z_interp = self.evaluate_bivariate_spline(coef)
        return Z_interp
